import datetime

import numpy as np
import pandas as pd
import pyqtgraph as pg
from PySide2.QtCore import QObject, Qt, QTime, Slot
from PySide2.QtUiTools import loadUiType
from PySide2.QtWidgets import (
    QButtonGroup,
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QMenu,
    QRadioButton,
    QVBoxLayout,
    QWidget,
    QWidgetAction,
)

from mad_gui.config import Config
from mad_gui.models.local import PlotData
from mad_gui.models.ui_state import MODES
from mad_gui.plot_tools.labels import BaseRegionLabel
from mad_gui.plot_tools.labels.base_label import BaseEventLabel
from mad_gui.plot_tools.plots.base_plot import BasePlot
from mad_gui.plot_tools.plots.sensor_plot_mode_handler import (
    AddModeHandler,
    BaseModeHandler,
    EditModeHandler,
    InvestigateModeHandler,
    RemoveModeHandler,
    SyncModeHandler,
)
from mad_gui.qt_designer import UI_PATH
from mad_gui.state_keeper import StateKeeper
from mad_gui.utils.helper import resource_path
from mad_gui.utils.model_base import BaseStateModel, Property
from typing import Callable, Dict, List, Optional, Type, Union

channel_selector_path = str(UI_PATH / "channel_selector.ui")
ui_path = resource_path(channel_selector_path)
if ".ui" in ui_path:
    try:
        ChannelSelector, _ = loadUiType(ui_path)
    except TypeError:
        try:
            import sys
            import os
            from pathlib import Path

            uic_path = Path(os.sep.join(sys.executable.split(os.sep)[:-1])) / "Scripts"
            sys.path.append(str(uic_path))
            ChannelSelector, _ = loadUiType(ui_path)
        except TypeError as e:
            raise FileNotFoundError(
                "Probably python did not find `pyside2-uic`. See "
                '"https://mad-gui.readthedocs.io/en/latest/troubleshooting.html#pyside2-uic-not-found" for more '
                "information"
            ) from e


elif ".py" in ui_path:
    from mad_gui.qt_designer.build.channel_selector import Ui_Form as ChannelSelector  # noqa


class TimeAxisItem(pg.AxisItem):
    """Class to show the time hh:mm:ss on x-channel instead of seconds since start.

    Parameters
    ----------
    start_time
        the time at which the session recording started

    Notes
    -----
    Please be aware that this assumes equal time deltas between all samples. This might lead to the fact,
    that for long datasets the actual time does not align with the our calculation from the samples. To give you a
    rough idea, in a dataset of 2h length, we are approximately off by 3-4 seconds.
    """

    def __init__(self, start_time: QTime, parent, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_time = start_time
        self.parent = parent

    def tickStrings(self, values: List, *args):  # noqa
        # note, that the received values are seconds
        ms = np.array(values) * 1000
        self.parent.configure_style()
        self.setLabel(text="Time", units="hh:mm:ss")
        return [self.start_time.addMSecs(value).toString("hh:mm:ss") for value in ms]


class SensorPlotState(BaseStateModel):
    plot_channels = Property([], dtype=list)
    mode: MODES = Property("investigate", dtype=str)


class SensorPlot(BasePlot):
    """Can be used to visualize IMU data and to create, delete and interact with plotted labels.

    Parameters
    ----------
    parent
        Parent widget or frame
    """

    MODE_HANDLERS: Dict[MODES, Type[BaseModeHandler]] = {
        "add": AddModeHandler,
        "investigate": InvestigateModeHandler,
        "edit": EditModeHandler,
        "remove": RemoveModeHandler,
        "sync": SyncModeHandler,
    }

    def __init__(
        self,
        plot_data: PlotData,
        initial_plot_channels=Optional[List[str]],
        start_time=Optional[datetime.datetime],
        label_classes=List[BaseRegionLabel],
        event_classes: Optional[List[BaseEventLabel]] = None,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(
            plot_data=plot_data,
            initial_plot_channels=initial_plot_channels,
            label_classes=label_classes,
            event_classes=event_classes,
            parent=parent,
        )
        self.start_time = start_time
        self.is_main_plot = False

        self._skip_snap_to = False
        self.state = SensorPlotState()
        self.state.plot_channels = initial_plot_channels or list(plot_data.data.columns)
        # We set the data once with auto range and all further updates don't update the range again.
        self._set_plot_data(
            self.plot_data.data,
            self.plot_data.sampling_rate_hz,
            self.state.plot_channels,
            fix_channels=False,
            start_time=self.start_time,
        )
        if len(plot_data.data) > 10000 and getattr(Config.settings, "AUTO_DOWNSAMPLE", False):
            self.plotItem.setDownsampling(auto=True)
        self.plotItem.setClipToView(True)
        self.state.bind(
            Slot(list)(
                lambda x: self._set_plot_data(
                    self.plot_data.data,
                    self.plot_data.sampling_rate_hz,
                    x,
                    fix_channels=True,
                    start_time=self.start_time,
                )
            ),
            "plot_channels",
            initial_set=False,
        )

        self.mode_handler = InvestigateModeHandler(self)
        self.state.bind(self._change_mode, "mode", initial_set=True)

        self._add_channel_selection_menu()
        StateKeeper.video_window_closed.connect(self.remove_video_cursor_line)

    def adapt_to_opening_video_window(self):
        if self.sync_info is not None:
            self.add_video_cursor_line()
            self.autoRange()
            StateKeeper.data_position_changed.connect(self.move_video_cursor_line)
            StateKeeper.data_position_changed.connect(self.set_graph_position)

    def _update_plotted_channels(self):
        submenus = self.getPlotItem().vb.menu.ctrl
        menu = next(iter([menu for menu in submenus if isinstance(menu, ChannelSelector)]))
        self.state.plot_channels = [
            cb.objectName() for cb in menu.channel_names.findChildren(QCheckBox) if cb.isChecked()
        ]
        Config.settings.CHANNELS_TO_PLOT = self.state.plot_channels

    def _add_channel_selection_menu(self):
        # adapted from pg.graphicsItems.ViewBox.channelCtrlTemplate
        submenu = QMenu()
        submenu.setTitle("Select channels")
        widget = QWidget()
        ui = ChannelSelector()
        ui.setupUi(widget)
        # open on hovering
        action = QWidgetAction(self.getPlotItem().vb.menu)
        action.setDefaultWidget(widget)
        submenu.addAction(action)
        self.channels_selection_button_group = QButtonGroup(parent=ui.channel_names_layout, objectName="channelsWidget")
        self.channels_selection_button_group.setExclusive(False)
        for channel in self.plot_data.data.columns:
            cb = QCheckBox(channel)
            cb.setCheckState(Qt.Checked if channel in self.state.plot_channels else Qt.Unchecked)
            cb.setObjectName(channel)
            self.channels_selection_button_group.addButton(cb)
            ui.channel_names_layout.addWidget(cb)

        self.channels_selection_button_group.buttonToggled.connect(self._update_plotted_channels)

        # add to plotwidget
        self.getPlotItem().vb.menu.axes.append(submenu)
        self.getPlotItem().vb.menu.ctrl.append(ui)
        self.getPlotItem().vb.menu.widgetGroups.append(submenu)
        self.getPlotItem().vb.menu.addMenu(submenu)

    def is_data_plotted(self):
        plotted_data = self.getPlotItem().listDataItems()
        if plotted_data and plotted_data[0].getData()[0] is not None:
            return True
        return False

    def _set_plot_data(
        self,
        data: pd.DataFrame,
        sampling_rate_hz: float,
        channels_to_plot: List[str],
        fix_channels: bool = False,
        start_time: Optional[datetime.time] = None,
    ):
        """Plot data into the widget. Will erase previous plotted data.

        Parameters
        ----------
        data
            Contains sensor data that should be plotted. The default channels to be plotted can be set in the config
            file :mod:`mad_gui.config` by setting the variable "CHANNELS_TO_PLOT", see also :ref:`Setting constants
            <channels to_plot>`.
        sampling_rate_hz
            Sampling frequency of the passed data in unit of Hertz
        fix_channels
            If true, it will not reset the x/y-channel to all available data but keep the current limits
            (x_min and x_max).
        start_time
            If you pass a start time, the x-channel will not show seconds from start of the recording but the time of
            the day
        """
        if data is None:
            return

        self._clear_data()
        ax_bottom = self.getAxis("bottom")
        ax_bottom.setLabel(text="time [seconds]")

        if not isinstance(data.index[0], int):
            raise TypeError(
                f"The index of the dataframe created by a loader must contain integers. However, "
                f"in this case it contained {type(data.index[0])}. You may change that by using "
                f"`df.reset_index(drop=True)` in your loaders `load_sensor_data`."
            )

        if start_time:
            start_time_qt = QTime(start_time.hour, start_time.minute, start_time.second)
            channel_items = {"bottom": TimeAxisItem(start_time_qt, orientation="bottom", parent=self)}
            self.setAxisItems(channel_items)

        if fix_channels:
            self.disableAutoRange()
        else:
            self.enableAutoRange()

        # make it responsive even for zoomed-in large datasets
        self.getPlotItem().setClipToView(True)

        for channel_name in channels_to_plot:
            # make sure we use the same color for one channel even if only few channels are plotted
            color_index = np.argmax([channel_name == item for item in data.columns])

            data_to_plot = data[channel_name]
            self._plot_channel(data_to_plot, sampling_rate_hz, hues=len(data.columns), color_index=color_index)

        self.autoRange()

    def _plot_channel(self, data_to_plot: pd.Series, sampling_rate_hz: float, hues: int, color_index: int):
        x_axis = data_to_plot.index / sampling_rate_hz

        if getattr(Config.settings, "NORMALIZE_DISPLAYED_DATA", False) is True:
            data_zero_mean = data_to_plot - data_to_plot.mean()
            data_to_plot = data_zero_mean / (data_zero_mean.max() - data_zero_mean.min())

        self.plot(
            x=x_axis,
            y=data_to_plot,
            pen=pg.mkPen(width=2, color=pg.intColor(index=color_index, hues=hues, sat=180)),
        )

    def _change_mode(self, new_mode: MODES):
        """Adapt tool tip text depending on mode and remove potentially plotted green line indicating a new event.

        Parameters
        ----------
        new_mode
            One of `add`, `edit`, `remove`, or `investigate`
        """
        # Deactivate old mode_handler:
        self.mode_handler.deactivate()
        self.mode_handler = self.MODE_HANDLERS[new_mode](sensor_plot=self)
        self.set_tooltip(new_mode)

        # On mode change, we sync the annotation state:
        self._sync_annotations()

    def _sync_annotations(self):
        for label_class in self.label_classes:
            self.plot_data.annotations[label_class.name].data = self._get_labels_from_plot(label_class)
        for event_class in self.event_classes:
            self.plot_data.annotations[event_class.name].data = self._get_events_from_plot(event_class)
        self.plot_data.annotations["events"].data = self._get_events_from_plot(BaseEventLabel)

    def set_tooltip(self, mode: MODES):
        tips = {
            "add": "Click to set position of start / event / end",
            "edit": "Hover over a line to be moved or click on a label to edit its type / details",
            "investigate": "Zoom by hovering over channels and scrolling",
            "remove": "Hover over label to remove and click left mouse",
            "sync": "Move the lines such that\n   - the green lines indicate moments at which all data streams are at "
            "the start position\n   - the red lines indicate moments at which all data streams are at the end "
            "position",
        }
        self.setToolTip(tips[mode])

    def clear_labels(self, label_class):
        for item in self.items():
            if type(item) == label_class:  # pylint: disable=unidiomatic-typecheck
                # we need this kind of type check since stride labels inherit from activity labels and thus would also
                # be selected if using `isinstance(...)`
                self.delete_item(item)

    def _clear_data(self):
        for item in self.items():
            if isinstance(item, pg.graphicsItems.PlotDataItem.PlotDataItem):
                self.delete_item(item)

    def delete_item(self, item: QObject):
        if self.event_labels:
            for event in self.event_labels.values():
                self.parent.removeItem(event)
        self.removeItem(item)

    def keyPressEvent(self, ev):  # noqa
        # Camelcase method overwrites qt method
        self.mode_handler.handle_key_press(ev)
        # This is important! As it will forward unhandled events to the parent!
        super().keyPressEvent(ev)

    def mousePressEvent(self, ev):  # noqa
        # Camelcase method overwrites qt method
        self.mode_handler.handle_mouse_click(ev)

    def mouseMoveEvent(self, ev):  # noqa
        # Camelcase method overwrites qt method
        self.mode_handler.handle_mouse_movement(ev)

    def _snap_to(self, pos: float, f: Callable):
        if self._skip_snap_to:
            return pos
        sampling_rate = self.plot_data.sampling_rate_hz
        pos_sample = pos * sampling_rate
        snap_min = int(pos_sample - getattr(Config.settings, "SNAP_RANGE_S", 0.1) / 2 * sampling_rate)
        snap_max = int(pos_sample + getattr(Config.settings, "SNAP_RANGE_S", 0.1) / 2 * sampling_rate)
        if not hasattr(Config.settings, "SNAP_CHANNEL"):
            ds = SelectSnapChannel(parent=self.parent, channels=list(self.plot_data.data.columns))
            ds.ask_user()
            snap_channel, remember = ds.get_snap_channel()
            if not snap_channel:
                return pos
            del ds
            if remember:
                Config.settings.SNAP_CHANNEL = snap_channel
        else:
            snap_channel = Config.settings.SNAP_CHANNEL
        region_data = self.plot_data.data[snap_channel].iloc[snap_min:snap_max]
        values = region_data.values
        idx_relative = f(values)
        idx_absolute = region_data.index[idx_relative]
        return idx_absolute / sampling_rate

    def snap_to_max(self, pos: float):
        """Snap to a maximum in the `Config.settings.SNAP_CHANNEL` of the sensor.

        Parameters
        ----------
        pos
            Position on the x-channel given in seconds

        Returns
        -------
        pos_snapped
            maximum within pos +- SNAP_RANGE_S (in seconds)/2 defined in your settings.
        """
        return self._snap_to(pos, np.argmax)

    def snap_to_min(self, pos: float):
        """Snap to a minimum in the `Config.settings.SNAP_CHANNEL` of the sensor.

        Parameters
        ----------
        pos
            Position on the x-channel given in seconds

        Returns
        -------
        pos_snapped
            minimum within pos +- SNAP_RANGE_S (in seconds)/2 defined in your settings.
        """
        return self._snap_to(pos, np.argmin)

    @staticmethod
    def _get_appropriate_stride_id():
        # TODO: find the proper stride number and renumber strides in stride_list
        return 0

    def _iter_labels_from_plot(self, label_type: Type[Union[BaseRegionLabel, BaseEventLabel]]):
        """Finds all instances of label_type in the plot and returns them."""
        for i_item in self.items():
            if type(i_item) is label_type:  # noqa
                # I need to use type, because StrideLabel inherits RegionLabel and thus I can not differentiate
                # them using `isinstance`
                yield i_item

    def _get_events_from_plot(self, label_type: Type[BaseEventLabel]) -> pd.DataFrame:
        events = self._iter_labels_from_plot(label_type)
        event_dicts = []
        for event in events:
            if event.belongs_to_region_label:
                continue
            pos = event.pos()
            data_dict = {
                "pos": int(pos.x() * self.plot_data.sampling_rate_hz),
                "min_height": event.span[0],
                "max_height": event.span[1],
                "description": event.description,
            }
            event_dicts.append(data_dict)
        df = pd.DataFrame.from_records(event_dicts)
        if df.empty:
            return df
        df.sort_values(by="pos", axis=0, inplace=True)
        return df.reset_index(drop=True)

    def _get_labels_from_plot(self, label_type: Type[BaseRegionLabel]) -> pd.DataFrame:
        """Finds all labels of type label_type and formats them into a pandas.DataFrame

        Parameters
        ----------
        label_type
            Some class, either RegionLabel or a class, which inherits from it.

        Returns
        -------
        labels
            A pandas.DataFrame with the columns label_id, start, end, type and details (see
            :class:`~mad_gui.plot_tools.RegionLabel`).
            In case of `label_type` being :class:`~mad_gui.plot_tools.StrideLabel`, there is an additional column `tc` (
            terminal contact).

        See Also
        --------
        :func:~mad_gui.plot_tools.SensorPlot._sync_annotations` will use this to synchronize the plotted labels with
        the labels in :class:`mad_gui.models.global_data.PlotData` object, which then can be accessed via the
        GUI's :class:`mad_gui.models.global_data.GlobalData`.
        """
        labels = self._iter_labels_from_plot(label_type)
        label_dicts = []
        for label in labels:
            start, end = label.getRegion()

            data_dict = {
                "identifier": label.id,
                "start": int(start * self.plot_data.sampling_rate_hz),
                "end": int(end * self.plot_data.sampling_rate_hz),
                "description": label.description,
            }
            for event_name, event_label in label.event_labels.items():
                data_dict.update({event_name: event_label.pos().x() * self.plot_data.sampling_rate_hz})
            label_dicts.append(data_dict)
        df = pd.DataFrame.from_records(label_dicts)
        if df.empty:
            return df
        df.sort_values(by="start", axis=0, inplace=True)
        return df.reset_index(drop=True)


class SelectSnapChannel(QDialog):
    def __init__(self, parent=None, channels=None):
        super().__init__(parent)
        self.parent = parent
        self.setWindowModality(Qt.ApplicationModal)
        self.setPalette(self.parent.palette())
        self.vbox = QVBoxLayout()
        self.channels = channels
        self.add_checkboxes_to_vbox()
        self.add_ok_btn()
        self.remember = QCheckBox("Remember until closing GUI")
        self.remember.setStyleSheet(self.parent.ui.btn_add_label.styleSheet().replace("QPushButton", "QCheckBox"))
        self.vbox.addWidget(self.remember)
        self.setLayout(self.vbox)
        self.setWindowTitle("Select channel to use for snapping")

    def add_checkboxes_to_vbox(self):
        self.radio_buttons = {}
        for channel in self.channels:
            self.radio_buttons[channel] = QRadioButton(channel)
            self.radio_buttons[channel].setChecked(Qt.Unchecked)
            style = self.parent.ui.btn_add_label.styleSheet()
            self.radio_buttons[channel].setStyleSheet(style.replace("QPushButton", "QRadioButton"))
            self.vbox.addWidget(self.radio_buttons[channel])

    def get_snap_channel(self):
        for channel in self.channels:
            if self.radio_buttons[channel].isChecked():
                return self.radio_buttons[channel].text(), self.remember.isChecked()
        return None, None

    def add_ok_btn(self):
        self.ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        button_style = self.parent.ui.btn_add_label.styleSheet()
        self.ok_btn.setStyleSheet(button_style)
        self.ok_btn.clicked.connect(self.handle_ok_click)
        self.vbox.addWidget(self.ok_btn)

    def handle_ok_click(self):
        self.close()

    def ask_user(self):
        self.exec_()



import numpy as np
import pandas as pd
import pyqtgraph as pg
from PySide2.QtGui import QBrush, QColor, QLinearGradient, Qt
from PySide2.QtWidgets import QWidget

from mad_gui.config import Config
from mad_gui.plot_tools.labels.base_label import BaseRegionLabel, InvalidStartEnd
from mad_gui.state_keeper import StateKeeper
from typing import Callable, Optional, Type


class SegmentedStrideLabel(BaseRegionLabel):
    """Stride labels that only have a start and an end.

    This class can for example be used to generate stride labels as suggested by Barth et al. [1]_. This label looks
    like an :class:`RegionLabel`, just that it is placed in the area of `self.min_height` to
    `self.max_height`. Furthermore, its start and end can snap to minima by setting STRIDE_SNAP_TO_MIN
    (see `docs/consts_example.md`).

    .. [1] Barth, J., Oberndorfer, C., Kugler, P., Schuldhaus, D., Winkler, J., Klucken, J., & Eskofier, B. (2013).
       Subsequence dynamic time warping as a method for robust step segmentation using gyroscope signals of daily life
       activities. Proceedings of the Annual International Conference of the IEEE Engineering in Medicine and Biology
       Society, EMBS, 6744–6747. https://doi.org/10.1109/EMBC.2013.6611104

    See Also
    --------
    :class:`RegionLabel`

    """

    def __init__(self, identifier: int, start: int, end: int, parent, **_kwargs):
        super().__init__(identifier, start, end, parent)
        self.min_height = 0.2
        self.max_height = 0.8
        self.span = (self.min_height, self.max_height)
        self.configure_children()
        self.standard_brush = StrideLabel._generate_brush(self)  # --> RegionLabel
        self.setBrush(self.standard_brush)


class StrideLabel(BaseRegionLabel):
    """Stride labels that have a start, an end, and a terminal contact event.

    Parameters
    ----------
    events
        A pandas series, where the indexes are the event names (like for example `tc`) and the values are the event
        position in samples. Up to now we only support a the event `tc`.

    See Also
    --------
    :class:`RegionLabel`

    """

    def __init__(
        self,
        identifier: int,
        start: int,
        end: int,
        parent,
        events: Optional[pd.Series] = None,
        description: Optional[str] = None,
        details: Optional[str] = None,
    ):
        # TODO: transform tc into a list of gait events
        super().__init__(
            start=start, end=end, parent=parent, identifier=identifier, description=description, details=details
        )
        self.min_height = 0.25
        self.max_height = 0.75
        self.span = (self.min_height, self.max_height)
        self.configure_children()
        self.tc_color = Config.theme.FAU_COLORS["dark_blue"]
        if events is not None:
            tc = events.tc
            if np.isnan(tc):
                tc = end - 0.1 * self.parent.sampling_rate_hz
                self.tc_color = QColor("red")
            self._add_tc_event(tc / self.parent.sampling_rate_hz)
        # make data lines stay in foreground
        self.setFlag(self.ItemStacksBehindParent, True)
        self.standard_brush = self._generate_brush()
        self.setBrush(self.standard_brush)

    def _generate_brush(self):
        grad = QLinearGradient(self.getRegion()[0], 0, self.getRegion()[1], 0)
        grad.setColorAt(0, QColor(0, 0, 255, 100))
        grad.setColorAt(0.6, QColor(0, 0, 255, 50))
        grad.setColorAt(1, QColor(0, 0, 255, 0))
        return QBrush(grad)

    def _add_tc_event(self, tc_pos: float):
        """Adding a terminal contact to the stride label visualization.

        Parameters
        ----------
        tc_pos
            x-coordinate in seconds

        # TODO: Change signature s.t. it accepts an event to be consistent in SensorPlot._add_label_mouse_move_event()
        """
        tc = pg.InfiniteLine(
            pos=tc_pos,
            movable=True,
            pen=pg.mkPen(style=Qt.DashLine, color=QColor(0, 0, 0, 255)),
        )
        tc.span = (
            self.min_height,
            self.max_height,
        )
        stride_region = self.getRegion()
        tc.setParentItem(self)
        tc.setBounds(stride_region)
        self.lines.append(tc)
        self.configure_children()

    def _tc_pos_changed(self):
        fs = self.parent.sampling_rate_hz
        snapped_position = round(self.tc.pos()[0] * fs) / fs
        self.tc.setPos(snapped_position)
        StateKeeper.set_has_unsaved_changes(True)

    def _left_mouse_click_event(self, ev):
        if self.removable and ev.button() == Qt.LeftButton:
            self.parent.removeItem(self)

    def _region_changed(self):
        BaseRegionLabel._region_changed(self)
        stride_region = self.getRegion()
        self.lines[2].setBounds((stride_region[0] + 0.03, stride_region[1] - 0.03))
        self.setBrush(self._generate_brush())


# TODO: Change so that start and end are in the same coordinate system as the other labels
class PartialLabel(pg.LinearRegionItem):
    """Wrap a label instance during creation"""

    def __init__(
        self,
        raw_start,
        sampling_rate_hz: float,
        label_class: Type[BaseRegionLabel],
        label_id: int,
        label_parent: Optional[QWidget] = None,
        post_process: Optional[Callable] = None,
    ):
        self.sampling_rate_hz = sampling_rate_hz
        self.label_class = label_class
        self.label_id = label_id
        self.label_parent = label_parent
        self.post_process = post_process
        self.events = []
        start = raw_start
        if post_process is not None:
            start = post_process(raw_start)
        super(PartialLabel, self).__init__(  # pylint: disable=super-with-arguments
            values=(start, start),
            pen=pg.mkPen(Config.theme.FAU_COLORS["dark_blue"]),
            swapMode="block",
        )
        self.span = (label_class.min_height, label_class.max_height)
        for i_child in self.childItems():
            i_child.span = self.span
        self.line_style = Qt.SolidLine
        if self.label_class is StrideLabel:
            self.line_style = Qt.DashLine

        self.setMovable(False)
        self.sigRegionChangeFinished.connect(self.on_range_change)

    def on_range_change(self):
        # Adapt color based on label class
        color = QColor(0, 255, 0)
        children = self.childItems()
        children[-1].setPen(pg.mkPen(color=color, style=self.line_style, width=2))

    def update_end(self, new_end):
        current_region = self.getRegion()
        new_region = (current_region[0], new_end)
        self.setRegion(new_region)

    def finalize(self):
        start, end = self.getRegion()
        if self.post_process is not None:
            end = self.post_process(end)

        if start == end:
            raise InvalidStartEnd()
        events_series = pd.Series()
        for event in self.events:
            s = pd.Series(data=event.pos().x() * self.sampling_rate_hz, index=event.description)
            events_series = events_series.append(s)
        final_label = self.label_class(
            identifier=self.label_id,
            start=start * self.sampling_rate_hz,
            end=end * self.sampling_rate_hz,
            parent=self.label_parent,
            events=events_series,
        )
        if self.label_class.descriptions:
            # If it is an activity label, we need to select the activity type
            final_label.edit_label_description()
        return final_label


class SynchronizationLabel(BaseRegionLabel):
    def __init__(self, start, end, parent=None):
        super().__init__(start=start, end=end, parent=parent)

    def _set_border_colors(self, start, end):
        self.start_color = QColor(0, 255, 0, 255)
        self.end_color = QColor("red")

    def _region_changed(self):
        return

    def _left_mouse_click_event(self, ev):
        """Synchronization labels are not clickable."""

    def _hover_event(self, ev):
        """We do not need to show anything since it is clear that it is a sync label."""



import warnings

import pandas as pd
from PySide2.QtCore import QObject, Qt, QUrl
from PySide2.QtMultimedia import QMediaContent, QMediaPlayer, QMediaPlaylist

from mad_gui.components.dialogs import UserInformation
from mad_gui.qt_designer.ui_video import UiVideoWindow
from mad_gui.state_keeper import StateKeeper


class VideoWindow(UiVideoWindow, QObject):
    """Display a video that can be synchronised with sensor data.

    Interacts with :class:`mad_gui.plot_tools.SensorPlot`"""

    def __init__(self, parent=None):  # MainWindow
        super().__init__()
        self.parent = parent
        if parent:
            self.setWindowIcon(self.parent.windowIcon())
            self.setPalette(self.parent.palette())
        self.fps = None
        self.sync_info = None
        self.duration = None
        self.video_file = None
        self.slider.sliderPressed.connect(self.slider_pressed)
        self.slider.sliderReleased.connect(self.slider_released)
        self.slider.sliderMoved.connect(self.slider_moved)
        self.slider.valueChanged.connect(self.slider_moved)
        self.player.positionChanged.connect(self.frame_changed)
        self.player.durationChanged.connect(self.set_slider_range)
        self.btn_play_pause.clicked.connect(self.toggle_play)
        self.setStyleSheet(parent.styleSheet())
        self.setWindowFlag(Qt.WindowStaysOnTopHint)
        self.user_informed_about_error = False
        self._times_set_rate_called = 0

    def _init_position(self):
        """Move the window to the center of the parent window."""

        x = self.parent.pos().x() + self.parent.size().width() / 2 - self.size().width() / 2
        y = self.parent.pos().y() + self.parent.size().height() / 2 - self.size().height() / 2
        self.move(x, y)

    def toggle_play(self):
        if self.player.state() == QMediaPlayer.PlayingState:
            self.player.pause()
        else:
            self.player.play()

    def start_video(self, video_file: str):
        self.video_file = video_file
        self.playlist.clear()
        self.playlist.addMedia(QMediaContent(QUrl.fromLocalFile(video_file)))
        self.playlist.setCurrentIndex(0)
        self.playlist.setPlaybackMode(QMediaPlaylist.CurrentItemInLoop)
        self.player.setPlaylist(self.playlist)
        self.player.mediaStatusChanged.connect(self.set_rate)
        self.player.setVideoOutput(self.view_video)
        self.player.setNotifyInterval(10)
        self.player.setPlaybackRate(1.0)
        self.player.setVolume(0)
        # trigger mediaStatusChanged event, such that self.fps is set in self.set_rate
        self.player.play()
        self.player.pause()
        self._init_position()

    def set_rate(self):
        if self.fps:
            # the signal that calls this will occasionally be called during playing the video, but we simply assume
            # that fps is constant
            return
        if "VideoFrameRate" not in self.player.availableMetaData():
            self._obtain_frame_rate_using_vlc()
        else:
            self.fps = self.player.metaData("VideoFrameRate")
            self.duration = self.player.metaData("Duration")

        if self.fps is None or self.duration is None:
            warnings.warn("Video duration or fps unknown.")
            return

        # get video duration in ms
        StateKeeper.video_duration_available.emit(self.duration / 1000, self.fps)

        # Not sure yet why this is, but we need the following commands to make sure switching to sync mode directly
        # after loading the video works
        self.player.play()
        self.player.pause()

    def _obtain_frame_rate_using_vlc(self):
        try:
            import vlc  # pylint: disable=import-outside-toplevel
        except ModuleNotFoundError:
            self.user_informed_about_error = True
            UserInformation.inform(
                "Cannot obtain the framerate of the video, which is necessary for synchronizing. "
                "Possibly this can be fixed by installing VLC Media Player.\n"
                "Click the Learn More link below to get to the website.",
                help_link="https://www.videolan.org/vlc/index.de.html",
            )
            return
        player_vlc = vlc.MediaPlayer()
        player_vlc.set_mrl(self.video_file)
        vlc.libvlc_media_parse(player_vlc.get_media())
        self.fps = player_vlc.get_fps()
        self.duration = self.player.duration()
        del player_vlc

    def set_sync(self, start_frame: float, end_frame: float):
        self.sync_info = pd.Series(data=[start_frame, end_frame], index=[["start", "end"]])

    def key_press_event(self, **args):  # noqa (unused argument)
        if self.player.state() == QMediaPlayer.PlayingState:
            self.player.pause()
        elif self.player.state() == QMediaPlayer.PausedState:
            self.player.play()

    def slider_pressed(self):
        self.player.pause()
        self.slider_is_pressed = True

    def slider_released(self):
        self.set_video_position()
        self.slider_is_pressed = False

    def slider_moved(self):
        if self.player.state() == QMediaPlayer.PausedState:
            self.set_video_position()
            self.frame_changed()

    def set_slider_range(self, duration):
        self.slider.setRange(0, duration)

    def set_slider_position(self):
        if not self.slider_is_pressed:
            self.slider.setValue(self.player.position())

    def set_video_position(self):
        if self.player.state() == QMediaPlayer.PausedState:
            self.player.setPosition(self.slider.value())

    def frame_changed(self):
        if self.player.mediaStatus() == QMediaPlayer.MediaStatus.LoadedMedia or self.duration is None:
            warnings.warn("Video is not playing or duration unknown.")
            return
        if not self.player.state() == QMediaPlayer.PausedState:
            self.set_slider_position()
        # if self.sync_info is not None and self.fps:
        #    stream_length = self.sync_info["end"][0] - self.sync_info["start"][0]
        #    start = self.sync_info["start"][0]
        # else:
        if self.sync_info is None:
            start = 0
            end = self.duration
        else:
            start = self.sync_info["start"] / self.fps * 1000
            end = self.sync_info["end"] / self.fps * 1000

        stream_length = end - start
        pos = self.slider.value()
        percent_since_start = (pos - start) / stream_length * 100
        try:
            StateKeeper.data_position_changed.emit(percent_since_start)
        except RuntimeError:
            # application already closed
            pass

    def closeEvent(self, event):  # noqa
        # comment "noqa" suppresses pep8 error N802: function name 'closeEvent' should be lowercase
        self.player.pause()
        StateKeeper.video_window_closed.emit()



"""This module keeps a class to create plots with IMU data and annotations. The annotations are kept in a separate
class."""

import numpy as np

from mad_gui.models.local import PlotData
from mad_gui.models.ui_state import MODES
from mad_gui.plot_tools.labels.base_label import BaseRegionLabel
from mad_gui.plot_tools.plots.base_plot import BasePlot
from mad_gui.plot_tools.plots.sensor_plot import SensorPlotState
from mad_gui.plot_tools.plots.sensor_plot_mode_handler import BaseModeHandler, InvestigateModeHandler
from mad_gui.plot_tools.plots.video_plot_mode_handler import SyncModeHandler
from mad_gui.state_keeper import StateKeeper
from typing import Dict, List, Optional, Type


class VideoPlot(BasePlot):
    """A graph of this class will be shown in the main window in order to synchronize video data with sensor data."""

    MODE_HANDLERS: Dict[MODES, Type[BaseModeHandler]] = {
        "investigate": InvestigateModeHandler,
        "edit": InvestigateModeHandler,
        "remove": InvestigateModeHandler,
        "add": InvestigateModeHandler,
        "sync": SyncModeHandler,
    }

    def __init__(self, parent=None, video_window=None):
        super().__init__(label_classes=[BaseRegionLabel], parent=parent)
        StateKeeper.video_duration_available.connect(self.update_video_duration)
        self.state = SensorPlotState()
        self.mode_handler = InvestigateModeHandler(self)
        self.state.bind(self._change_mode, "mode", initial_set=False)
        self.video_window = video_window

    def move_video_cursor_line(self, percent_since_start: float):
        self.video_cursor_line.setValue(self.video_window.slider.value() / 1000)

    def _change_mode(self, new_mode: MODES):
        """Adapt tool tip text depending on mode and remove potentially plotted green line indicating a new event.

        Parameters
        ----------
        new_mode
            One of `add`, `edit`, `remove`, or `investigate`
        """
        # Deactivate old mode_handler:
        self.mode_handler.deactivate()
        self.mode_handler = self.MODE_HANDLERS[new_mode](self)
        self._set_tooltip(new_mode)

    def distribute_video_sync(self):
        # Better: we should bind the sync items region to the sync of the VideoWindow
        self.video_window.start = self.sync_item.getRegion()[0] * 1000  # needs it in milli-seconds
        self.video_window.end = self.sync_item.getRegion()[1] * 1000

    def _set_tooltip(self, mode: MODES):
        tips = {
            "sync": "Move the lines such that\n   - the green lines indicate moments at which all data streams are at "
        }
        tooltip = tips.get(mode, None)
        self.setToolTip(tooltip)

    def update_video_duration(self, length_seconds: float, fps: float):
        percentage = np.asarray([float(n) for n in range(0, 101)])
        x_values = percentage / 100 * length_seconds
        self.set_data(x=x_values, y=np.zeros(len(x_values)), fps=fps)

    def set_data(self, x: List, y: List, fps: Optional[float] = 1):
        self.plot(x=x / 1000, y=y)
        self.plot_data = PlotData(data=y, sampling_rate_hz=fps)
        ax_bottom = self.getAxis("bottom")
        if fps != 1:
            ax_bottom.setLabel(text="time [seconds]")
        else:
            ax_bottom.setLabel(text="time [samples]")

    def add_sync_item(self):
        # just make sure we have the correct sampling frequency and video duration
        self.update_video_duration(self.video_window.player.duration(), self.video_window.fps)
        super().add_sync_item()



# pylint: skip-file
"""The :py:mod:`mad_gui.consts` keeps different configuration classes.
"""
from typing import Type

from mad_gui.config.settings import BaseSettings
from mad_gui.config.theme import BaseTheme


class _Config:
    """See docs `Adjusting constants <https://mad-gui.readthedocs.io/en/latest/README.html#adjusting-constants>`."""

    settings: Type[BaseSettings]
    theme: Type[BaseTheme]

    def set_theme(self, theme: Type[BaseTheme]):
        if not issubclass(theme, BaseTheme):
            raise ValueError("A new theme must subclass the default theme")
        self.theme = theme

    def set_settings(self, settings: Type[BaseSettings]):
        if not issubclass(settings, BaseSettings):
            raise ValueError("A custom settings object must subclass the BaseSettings.")
        self.settings = settings


Config = _Config()

__all__ = ["Config", "BaseTheme", "BaseSettings"]



import argparse
import sys
from pathlib import Path

from PySide2.QtWidgets import QApplication

from mad_gui.config import BaseSettings, BaseTheme
from mad_gui.plot_tools.labels import BaseEventLabel, BaseRegionLabel
from mad_gui.plugins.base import BasePlugin
from mad_gui.plugins.example import (
    EnergyCalculator,
    ExampleExporter,
    ExampleImporter,
    StationaryMomentsDetector,
    ActivityLabel,
    Stride,
    MyEvent,
)
from mad_gui.windows import MainWindow
from typing import Optional, Sequence, Type


def start_gui(
    data_dir=Path("."),
    plugins: Optional[Sequence[BasePlugin]] = (
        ExampleImporter,
        StationaryMomentsDetector,
        EnergyCalculator,
        ExampleExporter,
    ),
    labels: Optional[Sequence[BaseRegionLabel]] = (ActivityLabel, Stride),
    events: Optional[Sequence[BaseEventLabel]] = (MyEvent,),
    settings: Optional[Type[BaseSettings]] = BaseSettings,
    theme: Optional[Type[BaseTheme]] = BaseTheme,
):
    """Use this function to start the GUI and pass your plugins, like importers and algorithms to it.

    Please look at the `See Also` section below, to learn more about how to create objects of the classes you can
    pass to this function.

    Parameters
    ----------
    events
        The event labels you want to use. They inherit from :class:`~mad_gui.plot_tools.labels.BaseEventLabel`.
    data_dir
        The base path where there user will be directed to when opening a file.
    plugins
        Mostly you will adapt the GUI via this. The plugins must inherit from either of these:
        :class:`~mad_gui.plugins.BaseImporter`, :class:`~mad_gui.plugins.BaseExporter`,
        :class:`~mad_gui.plugins.BaseAlgorithm`.
    labels
        The region labels you want to use. They inherit from :class:`~mad_gui.plot_tools.labels.BaseRegionLabel`.
    settings
        Change some settings like the snap range, if you set your label's `snap_to_min` or `snap_to_max` to `True`
    theme
        Change the two main colors (light and dark) of the GUI using this


    """
    # Create the Qt Application
    app = QApplication(sys.argv)
    form = MainWindow(
        parent=app, data_dir=data_dir, settings=settings, theme=theme, plugins=plugins, labels=labels, events=events
    )
    form.show()

    sys.exit(app.exec_())


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir")
    args = parser.parse_args()
    start_gui(args.data_dir)



from mad_gui.plugins.base import BasePlugin
from typing import List, Type, TypeVar

T = TypeVar("T", bound=BasePlugin)


def filter_plugins(plugin_list: List[Type[BasePlugin]], baseclass: Type[T]) -> List[Type[T]]:
    return [b for b in plugin_list if issubclass(b, baseclass)]



from pathlib import Path

from PySide2.QtCore import QCoreApplication
from PySide2.QtWidgets import QFileDialog

from typing import Optional


def ask_for_file_name(base_dir: Path, parent=None, file_type="*.*") -> Optional[str]:
    data_file = QFileDialog(parent=parent).getOpenFileName(
        parent=None, caption="Select file to open", dir=str(base_dir), filter=file_type
    )[0]
    if data_file == "":
        # User clicked cancel
        return None
    return data_file


def set_cursor(window, cursor_type):
    window.setCursor(cursor_type)
    QCoreApplication.processEvents()  # On windows, we have to force GUI to update the cursor



import os
import tempfile
import warnings
from datetime import datetime, timedelta
from pathlib import Path

from PySide2.QtCore import QObject, Signal, Slot

from typing import Callable, Optional


def set_and_bind_property(slot: Slot, model_class: QObject, property_name: str, initial_set: bool = True):
    if initial_set:
        value = getattr(model_class, property_name)
        slot(value)
    get_property_signal(model_class, property_name).connect(slot)


def set_and_bidirectional_bind(
    slot: Slot,
    signal: Signal,
    model_class: QObject,
    property_name: str,
    initial_set: bool = True,
    transformer: Optional[Callable] = None,
) -> Callable:
    set_and_bind_property(slot, model_class, property_name, initial_set)
    if transformer is not None:
        func = lambda *x: setattr(model_class, property_name, transformer(*x))  # noqa
        # noqa because otherwise we need nested methodes and that's also not so nice...
    else:
        func = lambda x: setattr(model_class, property_name, x)  # noqa
    signal.connect(func)
    return lambda: signal.disconnect(func)


def get_property_signal(model_class: QObject, property_name: str) -> Signal:
    return getattr(model_class, signal_attribute_name(property_name))


def signal_attribute_name(property_name):
    """Return a magic key for the attribute storing the signal name."""
    return f"{property_name}_changed"


def value_attribute_name(property_name):
    """Return a magic key for the attribute storing the property value."""
    return f"_{property_name}_prop_value_"


def resource_path(relative_path):
    """Get absolute path to resource, works for dev and for PyInstaller"""

    try:
        base_path, relative_path = _get_resource_path(relative_path)
    except (IndexError, FileNotFoundError):
        base_path = os.path.abspath(".")
    return os.path.join(base_path, relative_path)


def _get_resource_path(relative_path):

    paths = sorted(Path(tempfile.gettempdir()).iterdir(), key=os.path.getctime)[::-1]
    mei_paths = [p for p in paths if "_MEI" in p.name]
    time_now = datetime.today()
    newest_mei = mei_paths[0]
    date_newest_mei = datetime.fromtimestamp(os.stat(newest_mei).st_ctime)
    if (time_now - date_newest_mei) > timedelta(seconds=60):
        raise FileNotFoundError("Did not find a current _MEI folder in tmp.")
    base_path = newest_mei
    relative_path = str.replace(relative_path, ".ui", ".py")
    relative_path = str.replace(relative_path, "qt_designer", f"qt_designer{os.sep}build")
    warnings.warn(
        f"Found a _MEI folder in {tempfile.gettempdir()}, which has bee created <1 minute ago."
        f" Therefore, I assume this is called from a standalone executable."
        f" For this reason, I'm changing the resource path, from mad_gui/qt_designer/*.ui to"
        f" mad_gui/qt_designer/build/*.py"
    )

    return base_path, relative_path



import numpy as np
import pandas as pd
import pyqtgraph as pg
from pyqtgraph import mkPen
from PySide2.QtGui import QColor, QHoverEvent, QMouseEvent, Qt

from mad_gui.components.dialogs import UserInformation
from mad_gui.components.dialogs.label_annotation_dialog import NestedLabelSelectDialog
from mad_gui.config import Config
from mad_gui.state_keeper import StateKeeper
from typing import Optional


class InvalidStartEnd(Exception):
    pass


class NoLabelSelected(Exception):
    pass


class BaseEventLabel(pg.InfiniteLine):
    name = "Event Label"
    min_height = 0
    max_height = 1
    descriptions: dict = {}

    def __init__(
        self,
        parent,
        pos,
        description: Optional[str] = None,
        belongs_to_region_label: Optional[bool] = False,
        min_height: Optional[float] = 0,
        max_height: Optional[float] = 1,
    ):
        self.parent = parent
        self.removable = False
        self.belongs_to_region_label = belongs_to_region_label
        self.base_pen = mkPen(color="b", style=Qt.DashLine)
        self.description = description
        self.min_height = min_height or 0
        self.max_height = max_height or 1
        pos_seconds = pos / self.parent.plot_data.sampling_rate_hz
        super().__init__(
            pos=pos_seconds, span=(self.min_height, self.max_height), pen=mkPen(color="b", style=Qt.DashLine)
        )

    def hoverEvent(self, event: QHoverEvent):  # noqa: N802
        """Actions when hovering over the InfiniteLine`"""
        mode = self.parent.state.mode
        if mode in ["edit", "sync", "remove"]:
            if event.enter and self.movable:
                self.parent.setCursor(Qt.SizeHorCursor)
            if event.enter and self.removable:
                self.parent.setCursor(Qt.PointingHandCursor)
                self.setPen(mkPen(QColor(255, 0, 0)))
            if event.exit:
                self.parent.setCursor(Qt.ArrowCursor)
                self.setPen(self.base_pen)
        else:
            if event.enter:
                # Unfortunately, this does not work
                # https://mathematica.stackexchange.com/q/66074
                # self.setToolTip(", ".join(self.description))
                pass
            else:
                self.parent.set_tooltip("investigate")

    def mousePressEvent(self, event: QMouseEvent):  # noqa: N802
        if self.removable:
            self.parent.removeItem(self)
            self.parent.setCursor(Qt.ArrowCursor)
        elif self.movable:
            super().mousePressEvent(event)
            self.setPos(self.parent.snap_to_sample(self.pos().x()))

    def mouseDragEvent(self, ev):  # noqa: N802
        super().mouseDragEvent(ev)
        self.setPos(self.parent.snap_to_sample(self.pos().x()))

    def make_editable(self):
        self.removable = False
        self.setMovable(True)

    def make_removable(self):
        self.removable = True
        self.setMovable(False)

    def make_readonly(self):
        self.removable = False
        self.setMovable(False)


class BaseRegionLabel(pg.LinearRegionItem):
    """A label region label plotted in the graph.

    Parameters
    ----------
    identifier
        An id for the label. Usually we just give it an increasing number.
    description
        A list of strings, filled by the user via the NestedLabelDialog
    start
        Start time of the label in samples.
    end
        End time of the label in samples.
    events
         A pd.DataFrame, with columns that are names for events and rows that have the positions of the events in
         samples since start of the data stream.
    plot_events
         self.events are only plotted if the name of the regarding event is in this list.
    parent
        A :class:`mad_gui.plot_tools.BasePlot` object.

    Attributes
    ----------
    name
        A string that is used to represent the label type in the GUI.

    """

    min_height = 0
    max_height = 1
    color = [150, 150, 150, 100]
    name = "Base Label"
    descriptions = None
    snap_to_min = False
    snap_to_max = False

    def __init__(
        self,
        start: int,
        end: int,
        parent,
        events: Optional[pd.Series] = None,
        identifier: int = None,
        description: Optional[str] = None,
        **_kwargs,  # underscore to prevent pylint form triggering
    ):
        pg.LinearRegionItem.__init__(self)
        self.parent = parent
        self.setMovable(False)
        self._set_border_colors(start, end)
        self._set_border_positions(start, end)
        self.event_labels = {}
        if events is not None:
            self._set_events(events)
        self.configure_children()
        self.standard_brush = pg.mkBrush(QColor(*self.color))
        self.sigRegionChangeFinished.connect(self._region_changed)
        self._set_movable(True)
        if start == end:
            return
        self.id = identifier
        self.description = description
        self.configure_children()
        self.span = (self.min_height, self.max_height)

        # The stride itself as a whole should not be movable, just the single lines (children = InfiniteLines)
        self.setAcceptHoverEvents(True)

        self.setBrush(self.standard_brush)
        self.setHoverBrush(QColor(0, 255, 0, 50))
        self.mouseClickEvent = self._left_mouse_click_event
        self.removable = False
        self.editable = False
        self._set_movable(False)
        self.hoverEvent = self._hover_event
        if self.description:
            if isinstance(description, tuple):
                description = ", ".join(self.description)
            else:
                description = self.description
            self.setToolTip(f"{self.name}: {description}")
        self.setEnabled(False)

    def _set_events(self, events: pd.DataFrame):
        for event, pos in events.iteritems():
            if pos is None:
                return
            self.event_labels[event] = BaseEventLabel(
                parent=self.parent,
                pos=pos,
                description=event,
                min_height=self.min_height,
                max_height=self.max_height,
                belongs_to_region_label=True,
            )

    def _left_mouse_click_event(self, ev):
        if self.removable and ev.button() == Qt.LeftButton:
            for event in self.event_labels.values():
                self.parent.removeItem(event)
            self.parent.removeItem(self)
            StateKeeper.set_has_unsaved_changes(True)
        elif self.editable and ev.button() == Qt.LeftButton:
            if not self.descriptions:
                UserInformation.inform(
                    f"MaD GUI is not aware of descriptions for the class {self.name}. Thus, you can not edit the "
                    f"description.",
                    help_link="https://mad-gui.readthedocs.io/en/latest/troubleshooting.html#mad-gui-is-not-aware-of-"
                    "descriptions-for-the-class",
                )
                return
            self.edit_label_description()

    def _hover_event(self, ev):
        """Coloring if mouse hovers of the stride"""
        # such that only borders of this element will be movable
        # and not the one of the neighbour
        self.setEnabled(True)

        hover_color = self._get_hover_color()
        self.setHoverBrush(hover_color)

        if self.removable or self.editable:
            if ev.enter:
                self.setMouseHover(True)
            if ev.exit:
                self.setMouseHover(False)
                self.setHoverBrush(self.standard_brush)

        description = self._description_to_str()
        self.setToolTip(f"{self.name}: {description}")

    def _get_hover_color(self) -> QColor:
        if self.removable:
            return QColor(255, 0, 0, 50)
        return QColor(0, 255, 0, 50)

    def _description_to_str(self):
        if isinstance(self.description, tuple):
            return ", ".join(self.description)
        return self.description

    def edit_label_description(self):
        """Setting the type of the activity to one given in the consts file.

        Called by :meth:`mad_gui.plot_tools.SensorPlot._finish_adding_activity` or if the user clicks on the label while
        being in edit mode. The emitted signal is caught by :meth:`mad_gui.MainWindow.ask_for_label_type`.
        """
        # the activities should be set by passing a `Settings` object which inherits from mad_gui.config.BaseSettings
        # and has an attribute `ACTIVITIES`, see our developer guidelines for more information
        new_description = NestedLabelSelectDialog(parent=self.parent.parent).get_label(self.descriptions)
        if not new_description:
            raise NoLabelSelected("Invalid description selected for label")
        self.description = new_description

    def _set_removable(self, removable: bool):
        self.removable = removable
        if not removable:
            self.setBrush(self.standard_brush)

    def _set_editable(self, editable: bool):
        self.editable = editable

    def _set_movable(self, movable: bool):
        for i_child in self.childItems():
            i_child.setMovable(movable)

    def make_editable(self):
        self._set_removable(False)
        self._set_editable(True)
        self._set_movable(True)

    def make_removable(self):
        self._set_removable(True)
        self._set_editable(False)
        self._set_movable(False)

    def make_readonly(self):
        self._set_removable(False)
        self._set_editable(False)
        self._set_movable(False)

    def _set_border_positions(self, start, end):
        sampling_rate_hz = self.parent.plot_data.sampling_rate_hz
        if np.isnan(start):
            start = end - 1.5 * sampling_rate_hz
        if np.isnan(end):
            end = start + 1.5 * sampling_rate_hz
        self.setRegion(
            (
                start / sampling_rate_hz,
                end / sampling_rate_hz,
            )
        )

    def _set_border_colors(self, start, end):
        """Color-code start and end of the label.

        It may happen that either the start or the end of a label is `None`. This can happen for example,
        when :func:`~mad_gui.plugins.BaseImporter.load_activity_stride_annotation_list` (or actually your
        implementation of it) finds a start for some activities but no end. This happens for example in the context of
        gait analysis, when a swing phase is found but no preceding stance phase - then there is no label of the start
        of the stride, but a label for terminal contact and the end of the stride.

        Color codes:
        red: The border is `None` and the user has to set it before saving evaluating. In case the user does not do
        that the event gets assigned to the red position before exporting data.
        orange: Either the user has not adapted the label and does not need to adapt it since loading was successful
        or the user has not dragged the lines after adding the label.
        blue: The user has adapted  the label successfully.

        Parameters
        ----------
        start
            Start time in samples
        end
            End time in samples
        """
        color = QColor(*self.color).toHsl()
        darker_color = color.darker(150).toRgb()
        self.start_color = darker_color
        self.end_color = darker_color
        if np.isnan(start):
            self.start_color = QColor("red")
        if np.isnan(end):
            self.end_color = QColor("red")

    def configure_children(self):
        for n_child, child in enumerate(self.childItems()):
            # make cursor horizontal when hovering
            # Pylint is right, that there are other ways, but in the hover-border-event, we want to refer to `self`
            # as this regionlabel and not as the border of the regionlabel
            child.hoverEvent = lambda event: self._hover_border_event(event)  # pylint: disable=unnecessary-lambda
            child.span = (self.min_height, self.max_height)

            # color of event lines
            self._set_child_color(child, n_child)

            # make events (InfiniteLines) movable
            child.sigPositionChanged.connect(self._reposition_lines)
            child.setMovable(False)

    def _set_child_color(self, child, n_th_child: int):
        if n_th_child == 0:
            if self.start_color == QColor("red"):
                width = 2
            else:
                width = 1
            child.setPen(self.start_color, width=width)
            return
        if n_th_child == 1:
            if self.end_color in [QColor("red"), QColor("green")]:
                width = 2
            else:
                width = 1
            child.setPen(self.end_color, width=width)

    def _reposition_lines(self):
        fs = self.parent.plot_data.sampling_rate_hz
        for i_child in self.childItems():
            snapped_position = round(i_child.pos()[0] * fs) / fs
            i_child.setPos(snapped_position)

    def _region_changed(self):
        """Called as soon as user drags start / end of a stride"""
        # border might have been red in the beginning, set it to blue if user has corrected this
        for i_child in self.childItems():
            if i_child.pen.color() == QColor("red"):
                # This gait event was 'None' before and therefore marked red. Apparently the user shifted it,
                # so we think it is now at the correct position
                i_child.pen.setColor(Config.theme.FAU_COLORS["dark_blue"])
        # TODO: change brush
        StateKeeper.set_has_unsaved_changes(True)

    def _hover_border_event(self, event: QHoverEvent):
        """Actions when hovering over the child item of type `pyqtgraph.InfiniteLine`"""
        mode = self.parent.state.mode
        if mode in ["edit", "sync"]:
            if event.enter:
                self.setMouseHover(False)
                self.parent.setCursor(Qt.SizeHorCursor)
                self.setEnabled(False)
            if event.exit and self.mouseHovering:
                self.parent.setCursor(Qt.PointingHandCursor)
                self.setMouseHover(True)
            elif event.exit and not self.mouseHovering:
                self.parent.setCursor(Qt.ArrowCursor)
                self.setEnabled(True)



from mad_gui.plot_tools.labels import BaseRegionLabel
from mad_gui.plot_tools.plots.sensor_plot_mode_handler import BaseModeHandler
from mad_gui.state_keeper import StateKeeper


class SyncModeHandler(BaseModeHandler):
    def __init__(self, plot):
        super().__init__(plot)
        self.plot.remove_video_cursor_line()
        self.plot.add_video_cursor_line()
        self.plot.add_sync_item()
        self.plot.show()
        self.plot.video_window.player.positionChanged.connect(self.plot.move_video_cursor_line)
        for item in self.plot.items():
            if isinstance(item, BaseRegionLabel):
                item.make_editable()

    def deactivate(self):
        self.plot.distribute_video_sync()
        self.plot.finish_syncing()
        self.plot.hide()
        StateKeeper.save_sync.emit()
        self.plot.video_window.player.positionChanged.disconnect(self.plot.move_video_cursor_line)

        for item in self.plot.items():
            if isinstance(item, BaseRegionLabel):
                item.make_readonly()



import pyqtgraph as pg
from PySide2.QtGui import QCursor
from PySide2.QtWidgets import QApplication

from mad_gui.plot_tools.plots.base_plot import BasePlot


class BaseModeHandler:
    mode: str

    def __init__(self, plot: BasePlot):
        self.plot = plot
        # This makes sure that the correct plot has focus when entering a new mode and can accept keyboard inputs
        # TODO: Should this be handled here on an application level?
        pos = QCursor.pos()
        active_widget = QApplication.widgetAt(pos)
        if active_widget:
            active_widget.setFocus()
        self._active = True

    def handle_key_press(self, ev):
        """What exactly should be done if a key is pressed needs to be defined by children."""

    def handle_mouse_click(self, ev):
        pg.PlotWidget.mousePressEvent(self.plot, ev)

    def handle_mouse_movement(self, ev):
        pg.PlotWidget.mouseMoveEvent(self.plot, ev)

    def deactivate(self):
        self._active = False



from PySide2.QtGui import QColor


class BaseTheme:
    COLOR_DARK = QColor(0, 56, 101)
    COLOR_LIGHT = QColor(188, 202, 220)
    PLOT_BACKGROUND = COLOR_LIGHT
    FAU_COLORS = {
        "dark_blue": QColor(0, 56, 101),
        "medium_blue": QColor(144, 167, 198),
        "light_blue": QColor(221, 229, 240),
    }

    FAU_NATFAK_COLORS = {
        "dark": QColor(0, 155, 119),
        "medium": QColor(170, 207, 189),
        "light": QColor(229, 239, 234),
    }

    FAU_TECHFAK_COLORS = {
        "dark": QColor(152, 164, 174),
        "medium": QColor(210, 213, 215),
        "light": QColor(235, 236, 238),
    }

    FAU_PHILFAK_COLORS = {
        "dark": QColor(201, 147, 19),
        "medium": QColor(217, 198, 137),
        "light": QColor(243, 238, 223),
    }



from typing_extensions import Literal

from mad_gui.utils.model_base import BaseStateModel, Property

MODES = Literal["investigate", "sync", "add", "modify", "remove"]


class UiState(BaseStateModel):
    menu_collapsed = Property(True)


class PlotState(BaseStateModel):
    x_range = Property((0, 1))
    y_range = Property((0, 1))
    x_range_max = Property((0, 1))
    mode: MODES = Property("investigate", dtype=str)



from mad_gui.plot_tools.labels.labels import SegmentedStrideLabel, StrideLabel, SynchronizationLabel, PartialLabel
from mad_gui.plot_tools.labels.base_label import BaseRegionLabel, BaseEventLabel

__all__ = [
    "BaseRegionLabel",
    "BaseEventLabel",
    "StrideLabel",
    "SegmentedStrideLabel",
    "SynchronizationLabel",
    "PartialLabel",
]



# pylint: skip-file
"""The :py:mod:`mad_gui.windows` keeps files for creating a set of windows.
"""
from pathlib import Path

import sys

# make sure window_buttons_rc is found
sys.path.append(str((Path(__file__).parent.parent.parent / "mad_gui" / "qt_designer").absolute()))

from mad_gui.windows.video_window import VideoWindow
from mad_gui.components.dialogs.user_information import UserInformation
from mad_gui.windows.main import MainWindow


__all__ = [
    "MainWindow",
    "VideoWindow",
    "UserInformation",
]



from functools import reduce

from PySide2.QtCore import QEvent, Qt
from PySide2.QtWidgets import QButtonGroup, QDialog, QDialogButtonBox, QGroupBox, QHBoxLayout, QRadioButton, QVBoxLayout

from typing import Any, Dict, List, Tuple, Union


def depth(d):
    if isinstance(d, dict):
        return 1 + (max(map(depth, d.values())) if d else 0)
    return 0


class NestedLabelSelectDialog(QDialog):
    """A Window with Radio Buttons or Checkboxes"""

    _max_depth: int
    _label_options: Union[List[str], Dict[str, Any]]

    latest_selection_: Tuple[str, ...] = ()

    def __init__(self, parent=None):
        super().__init__()
        self.parent = parent
        self.setWindowIcon(self.parent.windowIcon())
        self.setPalette(self.parent.palette())
        self.main_layout = QVBoxLayout()
        self.level_widgets: List[QVBoxLayout] = []
        self.level_button_group: List[QButtonGroup] = []
        self.buttons = QDialogButtonBox(parent=self)
        self.buttons.addButton("Ok", QDialogButtonBox.AcceptRole)
        self.buttons.addButton("Cancel", QDialogButtonBox.RejectRole)
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)
        self.current_selection_ = NestedLabelSelectDialog.latest_selection_
        self.setWindowTitle("Set description")

        self.setLayout(self.main_layout)

    def _setup_layout(self):
        # Create emtpy layouts based max depth
        label_layout = QHBoxLayout()
        self.main_layout.addLayout(label_layout)
        for i in range(self._max_depth + 1):
            group_box = QGroupBox(parent=self, title=f"Level {i}")
            group_box.setPalette(self.palette())
            inner_layout = QVBoxLayout()
            inner_layout.addStretch()
            group_box.setLayout(inner_layout)
            label_layout.addWidget(group_box)
            self.level_widgets.append(inner_layout)
            button_group = QButtonGroup(parent=group_box)
            button_group.setObjectName(str(i))
            button_group.buttonToggled.connect(self._on_label_select)
            self.level_button_group.append(button_group)

        self.main_layout.addWidget(self.buttons)
        self._setup_level(self._get_level_keys(self._label_options), level=0)
        if NestedLabelSelectDialog.latest_selection_:
            for level, name_button in enumerate(NestedLabelSelectDialog.latest_selection_):
                radiobutton = self.findChild(QRadioButton, name_button)
                self.level_button_group[level].buttonToggled.emit(radiobutton, True)

        self.level_button_group[0].buttons()[0].setFocus()

    def _on_label_select(self, value):
        if not value:
            return None
        level = int(self.sender().objectName())
        choices = self._get_choices(level + 1)
        if choices is not None:
            return self._setup_level(choices, level + 1)
        # We have a final choice.
        # We do nothing, but we need to cleanup all further levels"
        for i in range(level + 1, self._max_depth + 1):
            self._clean_level(i)

        self.current_selection_ = tuple(b.checkedButton().objectName() for b in self.level_button_group[: level + 1])
        return None

    def _get_choices(self, level):
        if level == 0:
            choices = self._label_options
        else:
            choices = reduce(
                lambda x, y: x[y.checkedButton().objectName()],
                self.level_button_group[: level - 1],
                self._label_options,
            )
            if isinstance(choices, dict):
                choices = choices[self.level_button_group[level - 1].checkedButton().objectName()]
                if choices is None:
                    # We have a final choice
                    return None
            else:
                # We have a final choice
                return None
        return self._get_level_keys(choices)

    @staticmethod
    def _get_level_keys(level_choice: Union[List[str], Dict[str, Any]]):
        if isinstance(level_choice, dict):
            return list(level_choice.keys())
        return level_choice

    def _clean_level(self, level):
        group = self.level_button_group[level]
        for b in group.buttons():
            b.deleteLater()

    def _setup_level(self, choices: List[str], level: int):
        widget = self.level_widgets[level]
        group = self.level_button_group[level]

        self._clean_level(level)

        first_button = None

        for i, choice in enumerate(choices):
            button = QRadioButton(str(i + 1) + ": " + choice, parent=self)
            # This is required so that we can capture the space key
            button.installEventFilter(self)
            button.setObjectName(choice)
            if i == 0:
                first_button = button
            # We insert instead of add to keep the stretch at the bottom
            widget.insertWidget(widget.count() - 1, button)
            if (
                NestedLabelSelectDialog.latest_selection_
                and len(NestedLabelSelectDialog.latest_selection_) > level
                and NestedLabelSelectDialog.latest_selection_[level] in choice
            ):

                button.setChecked(True)
            group.addButton(button)

        # Enable first checkbox:
        if first_button is not None and not NestedLabelSelectDialog.latest_selection_:
            first_button.setChecked(True)

    def get_label(self, label_options: Union[List[str], Dict[str, Any]]):
        self._max_depth = depth(label_options)
        self._label_options = label_options
        self._setup_layout()
        if self.exec_():
            NestedLabelSelectDialog.latest_selection_ = self.current_selection_
            return self.current_selection_
        return None

    def keyPressEvent(self, event):  # noqa
        # Camelcase method overwrites qt method
        if event.key() in (Qt.Key_Return, Qt.Key_Space):
            event.accept()
            self.accept()
            return
        if event.key() == Qt.Key_Escape:
            event.accept()
            self.reject()
            return
        current_focus = self.focusWidget().parentWidget()
        level = self.level_widgets.index(current_focus.layout())
        choices = self._get_choices(level)
        # Number keys start at 49 to 59
        mapping = dict(zip(range(49, 59), choices))
        k_choice = mapping.get(event.key(), None)
        if k_choice is None:
            return
        button: QRadioButton = current_focus.findChild(QRadioButton, k_choice)
        button.setChecked(True)
        event.accept()

    def eventFilter(self, q_object, event) -> bool:  # noqa
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Space:
            self.keyPressEvent(event)
        return super(NestedLabelSelectDialog, self).eventFilter(q_object, event)  # noqa



from PySide2.QtCore import QSize, Qt
from PySide2.QtMultimedia import QMediaPlayer, QMediaPlaylist
from PySide2.QtMultimediaWidgets import QVideoWidget
from PySide2.QtWidgets import QHBoxLayout, QPushButton, QSizePolicy, QSlider, QVBoxLayout, QWidget


class UiVideoWindow(QWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Video")
        self.verticalLayout = QVBoxLayout()
        self.view_video = QVideoWidget()
        size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        size_policy.setHorizontalStretch(0)
        size_policy.setVerticalStretch(0)
        size_policy.setHeightForWidth(self.view_video.sizePolicy().hasHeightForWidth())
        self.view_video.setSizePolicy(size_policy)
        self.view_video.setMinimumSize(QSize(0, 200))
        self.view_video.setMaximumSize(QSize(1666666, 500))
        self.view_video.setSizeIncrement(QSize(0, 0))
        self.view_video.setBaseSize(QSize(0, 0))
        self.view_video.setFocusPolicy(Qt.NoFocus)
        self.view_video.setAcceptDrops(False)
        self.resize(876 // 2, 705 // 2)
        self.view_video.setAutoFillBackground(True)
        self.view_video.setStyleSheet("")
        self.view_video.setObjectName("view_video")
        self.verticalLayout.addWidget(self.view_video)

        self.player = QMediaPlayer()
        self.playlist = QMediaPlaylist()

        self.horizontalLayout = QHBoxLayout()

        self.btn_play_pause = QPushButton()
        self.btn_play_pause.setMaximumSize(QSize(150, 16777215))
        self.btn_play_pause.setObjectName("btn_play_pause")
        self.btn_play_pause.setText("Play / Pause")
        self.horizontalLayout.addWidget(self.btn_play_pause)

        self.slider = QSlider()
        self.slider.setOrientation(Qt.Horizontal)
        self.slider.setObjectName("slider")
        self.slider.setTickInterval(10)
        self.slider.setSingleStep(10)
        self.slider.setStyleSheet("QSlider::handle:horizontal{background-color: rgb(255,255,0);}")
        self.horizontalLayout.addWidget(self.slider)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.setLayout(self.verticalLayout)
        self.slider_is_pressed = False



from __future__ import annotations

import pandas as pd

from mad_gui.utils.model_base import BaseStateModel, Property
from typing import Dict, List, Optional


class AnnotationData(BaseStateModel):
    """An object that just represents a pandas.DataFrame and makes it possible to bind the dataframe to methods.

    This object is used to present starts, ends and descriptions of labels in a plot. It is used in
    :class:`mad_gui.models.local.PlotData.annotations`.

    Attributes
    ----------
    data
        An object which represents a pandas.DataFrame at least with the columns `start` and `end`.
        Each row encodes one label.

    Methods
    -------
    to_df
        Utility method to return self.data as a pandas.DataFrame.

    Examples
    --------
    >>> plot_data = gui.global_data.plot_data
    >>> for plot, data in plot_data.items():
    ...      print(f"Sensor {plot}'s annotations are:")
    ...      for label_class, annotation_df in data.annotations.items():
    ...           print(f"Labels of class {label_class}: ")
    ...           print(annotation_df.data)
    ...
    Sensor Pocket IMU's annotations are:
    Labels of class Base Label:
    identifier  start   end description details
             0   1687  2238    (sleep,)    None
    """

    data = Property(pd.DataFrame(), dtype=pd.DataFrame)

    def to_df(self):
        """Return the annotations stored in this object as :class:`pandas.DataFrame`"""
        return self.data


class PlotData(BaseStateModel):
    """An object, which keeps the plotted data and annotations of a single plot.

    Parameters
    ----------
    data
        A pandas.DataFrame, where each column is one channel of plotted data.

    sampling_rate_hz
        The sampling rate with which the data was recorded.

    annotation
        A dictionary, where the keys are the label names (as named in the label's
        :meth:`~mad_gui.plot_tools.labels.BaseRegionLabel.name` attribute). The values are instances of
        :class:`~mad_gui.models.local.AnnotationData`.

    additional_data
        Keeps things that belongs to the plotted data but should not be plotted. Here you can find everything that was
        returned from your loader for one sensor, where the key is not `sensor_data` or `sampling_rate_hz`, see
        :class:`mad_gui.plugins.BaseImporter`.
    """

    def __init__(self, data: pd.DataFrame, sampling_rate_hz: float, annotation: Dict = None, additional_data=None):
        self.data = data
        self.sampling_rate_hz = sampling_rate_hz
        self.annotations = annotation or {}
        self.additional_data = additional_data

    def to_dict(self):
        """Represent this object as a dictionary, such that it can be pickled.

        Returns
        -------
        dict
            A dictionary with the keys `sensor_data`, `annotations`, `events`, and `sampling_rate_hz`,
            where the first three are :class:`pandas.DataFrame` and the last is a float.
        """
        return {
            "sensor_data": self.data,
            "annotations": {k: v.to_df() for k, v in self.annotations.items() if k != "events"},
            "events": self.annotations["events"].to_df(),
            "sampling_rate_hz": self.sampling_rate_hz,
        }

    @classmethod
    def from_dict(cls, plot_data: Dict, selections: Optional[List] = None) -> PlotData:
        """Create an instance of this class from a dictionary.

        Parameters
        ----------
        plot_data
            A dictionary with a key `sensor_data`, which is a :class:`pandas.DataFrame`; a key `sampling_rate_hz`,
            which is a float`; optionally a key `annotations` containing a dictionary, where keys are label names (
            :attr:`mad_gui.plot_tools.labels.BaseRegionLabel.name`) and each value is a :class:`pandas.DataFrame` at
            least with the columns `start` and `end` and optional columns `description` and `identifier`.
        selections
            This is used to indicate which of the items in the passed dictionary plot_data should be plotted. We use it
            only when using the "reload displayed data" button.

        Returns
        -------
        PlotData
        """

        selections = selections or plot_data.keys() - {"sampling_rate_hz"}

        if not all(required_key in plot_data.keys() for required_key in ["sensor_data", "sampling_rate_hz"]):
            raise KeyError(
                "Your importer's `load_sensor_data` method must return a dict at least with the keys `sensor_data` "
                f"and `sampling_rate_hz`, but it has {plot_data.keys()}. For docstring on that method see https://mad-"
                "gui.readthedocs.io/en/latest/modules/generated/plugins/mad_gui.plugins.BaseImporter.html#mad_gui."
                "plugins.BaseImporter.load_sensor_data"
            )
        sensor_data = plot_data["sensor_data"]
        sampling_rate_hz = plot_data["sampling_rate_hz"]

        obj = cls(sensor_data, sampling_rate_hz)
        for selection in set(selections) - {"sensor_data", "sampling_rate_hz"}:
            if selection == "annotations":
                obj._add_annotations(plot_data)
                continue

            if not obj._add_label(plot_data, selection):
                obj._add_additional_data(obj, plot_data, selection)

        obj.annotations["events"] = AnnotationData()
        obj._add_events(getattr(plot_data, "events", None))
        return obj

    @staticmethod
    def _add_additional_data(obj, plot_data, selection):
        if not obj.additional_data:
            obj.additional_data = {}
        obj.additional_data[selection] = plot_data[selection]

    def _add_annotations(self, plot_data: Dict):
        if not plot_data.get("annotations", None):
            return
        for label in plot_data["annotations"].keys():
            self._add_label(plot_data, label)

    def _add_events(self, events: pd.DataFrame):
        if not events:
            return
        global_labels = [label for label in self.parent().global_data.labels if label.name in self.annotations]
        for _, event in events.iterrows():
            for label in global_labels:
                if event.min_height == label.min_height and event.max_height == label.max_height:
                    self.annotations["events"].data = self.annotations["events"].data.append(event)

    def _add_label(self, plot_data: Dict, label: str):
        if not plot_data.get("annotations", None):
            return False
        annotation = AnnotationData()
        if label in plot_data["annotations"].keys():
            # only in this case the plot_data object is aware of this kind of label
            annotation.data = plot_data["annotations"][label]
            self.annotations[label] = annotation
            return True
        return False



"""Keeps the state of the GUI and related signals / slots."""

from PySide2.QtCore import QObject, Signal, Slot


class _StateKeeper(QObject):
    """Used as a singleton to enable communication between objects that don't have a direct relationship.

    In the section `Other Parameters`, all Signals are listed. You can emit a certain signal by calling
    `StateKeeper.<signal name>.emit(<potential arguments, like a text>). Slots are listed as Methods and you can
    connect to them like this: `StateKeeper.<slot name>.connect(<the method you want to call upon emission of that
    signal>).

    Attributes
    ----------
    current_frame
        The frame that is shown by :class:`mad_gui.VideoWindow`
    label_state
        One of the states "add", "remove", "edit", or "investigate". State information is color-coded by the buttons
        in the
        upper part of the GUI. Can be changed by mouse-click or shortcuts "a", "e", or "Esc".
    gui_has_unsaved_changes
        Keeps information if there has been any user interaction since last time data was either exported using
        :meth:`mad_gui.MainWindow._export` or saved using :meth:`mad_gui.MainWindow._save_data_gui_format`.

    """

    announce_data_types = Signal(dict)  # MainWindow / DataSelector
    executed_algorithms = []

    save_sync = Signal()

    data_position_changed = Signal(float)
    video_window_closed = Signal()
    video_duration_available = Signal(float, float)
    gui_has_unsaved_changes = False
    plugins = []

    @Slot(int)
    def frame_changed(self, frame: int):
        self.signal_sample_changed.emit(self.synchronizer.frame_to_sample(frame))
        self.signal_frame_changed.emit(frame)

    @Slot(bool)
    def set_has_unsaved_changes(self, flag: bool):
        self.gui_has_unsaved_changes = flag


StateKeeper = _StateKeeper()



import os
from pathlib import Path

import numpy as np
import pandas as pd
from PySide2.QtWidgets import QFileDialog

from mad_gui.components.dialogs import UserInformation
from mad_gui.models import GlobalData
from mad_gui.models.local import PlotData
from mad_gui.plot_tools.labels import BaseRegionLabel, BaseEventLabel
from mad_gui.plugins.base import BaseAlgorithm, BaseExporter, BaseImporter
from typing import Dict

from mad_gui.state_keeper import StateKeeper


class ActivityLabel(BaseRegionLabel):
    min_height = 0.8
    max_height = 1
    name = "Activity"
    descriptions = {"Jump": None, "Walk": ["Slow", "Normal", "Fast"]}


class MyEvent(BaseEventLabel):
    min_height = 0
    max_height = 1
    name = "Peak"
    descriptions = {"Positive peak": None, "Negative peak": None}
    snap_to_min = False


class Stride(BaseRegionLabel):
    min_height = 0
    max_height = 0.75


class ExampleImporter(BaseImporter):
    """An exemplary importer.

    Attributes
    ----------
    file_type
        Can restrict the format of the data/video/annotation file that can be loaded with this importer.
    """

    file_type = {"data_file": "*.csv", "video_file": "*.mp4", "annotation_file": "*.csv"}

    @classmethod
    def name(cls) -> str:
        return "Example Importer"

    def load_sensor_data(self, file: str) -> Dict:
        df = pd.read_csv(file)
        data = {
            "Pocket IMU": {
                "sensor_data": df[["acc_x", "acc_y", "acc_z", "gyr_x", "gyr_y", "gyr_z"]],
                "sampling_rate_hz": 1 / df["time"].diff().mean(),
            }
        }
        return data


class StationaryMomentsDetector(BaseAlgorithm):
    """An exemplary algorithm, which uses the data to create annotations.

    Note: a label inherited from :class:`~mad_gui.plot_tools.BaseRegionLabel with the name `Activity` must have been
    passed to :meth:`~mad_gui.start_gui` in order for this algorithm to create annotations.
    """

    @classmethod
    def name(cls):
        return "Find Resting Phases (MaD GUI example)"

    def process_data(self, plot_data: Dict[str, PlotData]):
        for sensor_plot in plot_data.values():
            sensor_plot.annotations["Activity"].data = self.get_annotations(sensor_plot.data)

    @staticmethod
    def _get_standing_windows(data: pd.DataFrame, window_length: int):
        diff = data.diff().apply(np.linalg.norm, axis=1)

        standing_windows = abs(diff.rolling(window=window_length).mean()) < 0.2

        # from standing_windows we get only a single one if the complete windows is standing -> we need to transform
        # this single one to a series of window_length ones
        standing = np.zeros(shape=(1, len(data)))
        for idx, value in standing_windows.iteritems():
            if value:
                filter_lag = window_length
                start = max(0, idx - filter_lag)
                stop = min(idx + window_length - filter_lag, len(data))
                standing[0, start:stop] = 1
        return standing

    @staticmethod
    def _binary_to_df(array: np.ndarray):
        # now we have something like [0 1 1 1 1 0], which we want to transform to start: 1, end: 4
        df = pd.DataFrame(array).T
        starts_stops = df.diff()
        annotations = pd.DataFrame(columns=["start", "end"])
        start = stop = None
        for idx in df.index:
            if starts_stops[0].iloc[idx] == 1:
                start = idx
            if start and (starts_stops[0].iloc[idx] == -1 or idx == len(df)):
                stop = idx
            if start and stop:
                annotations = annotations.append(pd.DataFrame(data=[[start, stop]], columns=["start", "end"]))
                start = None
                stop = None
        return annotations

    def get_annotations(self, data: pd.DataFrame):
        acc = data[["acc_x", "acc_y", "acc_z"]]

        # sampling rate was 102.4 and we choose ~one second as window length
        flags_standing = self._get_standing_windows(acc, window_length=102)
        annotations = self._binary_to_df(flags_standing)
        annotations["description"] = "standing"
        return annotations


class EnergyCalculator(BaseAlgorithm):
    """An exemplary algorithm, which evaluates data within each annotation to calculate a feature.

    Note: a label inherited from :class:`~mad_gui.plot_tools.BaseRegionLabel with the name `Activity` must have been
    passed to :meth:`~mad_gui.start_gui`. Furthermore, annotations of this kind must exist in the GUI, for example
    created by :class:`mad_gui.plugins.example.StationaryMomentsDetector` in order for this algorithm to calculate
    features for each existing annotation.
    """

    @classmethod
    def name(cls):
        return "Mean energy of acceleration (MaD GUI example)"

    def process_data(self, plot_data: Dict[str, PlotData]):
        for sensor_plot in plot_data.values():
            for i_activity, activity in sensor_plot.annotations["Activity"].data.iterrows():
                description = sensor_plot.annotations["Activity"].data.at[i_activity, "description"]
                sensor_plot.annotations["Activity"].data.at[i_activity, "description"] = (
                    str(description)
                    + " ("
                    + self.calculate_features(sensor_plot.data.iloc[activity.start : activity.end])
                    + ")"
                )

    @staticmethod
    def calculate_features(sensor_data: pd.DataFrame) -> str:
        signal = sensor_data[["acc_x", "acc_y", "acc_z"]]
        energy = np.sqrt((signal**2).sum(axis=1).mean())
        return f"mean acceleration = {energy:.2f}"


class ExampleExporter(BaseExporter):
    """An exemplary exporter, which writes all existing annotations into a csv file."""

    @classmethod
    def name(cls):
        return "Export annotations to csv (MaD GUI example)"

    def process_data(self, global_data: GlobalData):
        directory = QFileDialog().getExistingDirectory(
            None, "Save .csv results to this folder", str(Path(global_data.data_file).parent)
        )
        for plot_name, plot_data in global_data.plot_data.items():
            for label_name, annotations in plot_data.annotations.items():
                if len(annotations.data) == 0:
                    continue
                save_file_name = (
                    directory + os.sep + plot_name.replace(" ", "_") + "_" + label_name.replace(" ", "_") + ".csv"
                )
                annotations.data.to_csv(save_file_name)
                algorithms_str = "Algorithms:" + ",".join([algo.__name__ for algo in StateKeeper.executed_algorithms])
                df_algorithms = pd.DataFrame(data=[algorithms_str])
                df_algorithms.to_csv(save_file_name, mode="a", header=False, index=False)

        UserInformation.inform(f"The results were saved to {directory}.")



"""A unique object per opened MaD GUI, keeping all relevant data."""

from __future__ import annotations

from mad_gui.models.local import PlotData
from mad_gui.utils.model_base import BaseStateModel, Property
from typing import Dict


class GlobalData(BaseStateModel):
    """A single object per MaD GUI instance that keeps global data.

    Attributes
    ----------
    active_loader
        One of the loaders passed to start_gui, which have inherited from :class:`~mad_gui.plugins.BaseImporter`. It
        can be selected by the user via the dropdown menu in the
        :class:`~mad_gui.components.dialogs.plugin_selection.LoadDataDialog`.
    annotation_file
        In case the user selected annotations when loading data in the
        :class:`~mad_gui.components.dialogs.plugin_selection.LoadDataDialog`, this keeps the file the user selected.
    data_file
        File of the data that is selected by the user in the
        :class:`~mad_gui.components.dialogs.plugin_selection.LoadDataDialog`.
    sync_file
        A file that keeps synchronization between video and sensor data. The GUI automatically searches for a file in
        the same folder as the video_file and if it finds a file, that has `*sync*.xlsx` in it assumes, this keeps
        the video synchronization.
    video_file
        File which contains the video to be displayed in the
        :class:`~mad_gui.components.dialogs.plugin_selection.LoadDataDialog`.
    plot_data
        A dictionary, where the keys are the names of the plots and the values are
        :class:`~mad_gui.models.local.PlotData` objects.
    plugins
        All plugins, that the GUI is aware of. The GUI is aware of all plugins that were passed to it via
        :meth:`~mad_gui.start_gui`.
    labels
        All the label classes that the GUI is aware of (= the ones that were passed to :meth:`~mad_gui.start_gui`).
        The actually plotted annotations can not be found here, but in `plot_data[<name of the
        plot>].annotations.data`,
        see the documentation of :class:`~mad_gui.models.local.PlotData`.

    """

    data_file = Property("", dtype=str)
    video_file = Property("", dtype=str)
    sync_file = Property("", dtype=str)
    annotation_file = Property("", dtype=str)
    base_dir = Property("", dtype=str)

    # TODO: make plot_data a working property, so we do not need to call _plot_data manually after using
    #  MainWindow.use_algorithm.
    plot_data: Dict[str, PlotData] = Property({}, dtype=dict)



"""A dialog which is used internally by the GUI to select data/video/annotation files and an importer."""

import traceback
import warnings
from pathlib import Path

import pandas as pd
from PySide2 import QtCore
from PySide2.QtGui import Qt
from PySide2.QtUiTools import loadUiType
from PySide2.QtWidgets import QDialog, QLabel, QLineEdit, QPushButton

from mad_gui import BaseImporter
from mad_gui.components.dialogs.user_information import UserInformation
from mad_gui.components.helper import ask_for_file_name, set_cursor
from mad_gui.config import Config
from mad_gui.qt_designer import UI_PATH
from mad_gui.utils.helper import resource_path
from mad_gui.utils.model_base import BaseStateModel, Property
from typing import Any, Dict, List, Optional, Tuple, Type

LINK_IMPLEMENT_IMPORTER = "https://mad-gui.readthedocs.io/en/latest/customization.html#implement-an-importer"

ui_path = resource_path(str(UI_PATH / "load.ui"))
if ".ui" in ui_path:
    try:
        LoadWindow, _ = loadUiType(ui_path)
    except TypeError:
        try:
            import sys
            import os

            uic_path = Path(os.sep.join(sys.executable.split(os.sep)[:-1])) / "Scripts"
            sys.path.append(str(uic_path))
            LoadWindow, _ = loadUiType(ui_path)
        except TypeError as e:
            raise FileNotFoundError(
                "Probably python did not find `pyside2-uic`. See "
                '"https://mad-gui.readthedocs.io/en/latest/troubleshooting.html#pyside2-uic-not-found" for more '
                "information"
            ) from e


elif ".py" in ui_path:
    from mad_gui.qt_designer.build.load import Ui_Form as LoadWindow  # noqa


class LoadDataDialogState(BaseStateModel):
    data_file = Property("", dtype=str)
    video_file = Property("", dtype=str)
    annotation_file = Property("", dtype=str)


class LoadDataDialog(QDialog):
    """A dialog, opened upon pressing `Load data` button in the GUI - interface to use importers.

    When the user wants to load data, they use this dialog. In the dialog, the user can use a dropdown menu to select an
    importer they want to use. This importer must previously have been passed to the GUI via the
    :meth:`mad_gui.start_gui`. The importer can be created as described in :ref:`implement importer`.

    Parameters
    ----------
    base_dir
        A base directory, which should be shown when the use wants to select a file - you can configure this by
        passing a `data_dir` to :meth:`mad_gui.start_gui`.
    loaders
        All the importers that were passed to :meth:`mad_gui.start_gui`, which will be shown in the dropdown.
    parent
        It is used to set the window and button style
    initial_state
        Will be assigned to `self.state`, which keeps the data of this view.

    Methods
    -------
    get_data
        After instantiating this class, use this method to open the dialog. Then, the user selects importer and data
        and this returns the loaded data in the MaD GUI format,
        see :meth:`mad_gui.plugins.BaseImporter.load_sensor_data`.
    validate_data_format
        Check whether the importer returned data in the expected formats and throw messages on what went wrong, if so.
    """

    final_data_: Dict[str, Dict[str, Any]]
    loader_: BaseImporter

    def __init__(
        self,
        base_dir: Path,
        loaders: List[Type[BaseImporter]],
        parent=None,
        initial_state: Optional[LoadDataDialogState] = None,
    ):
        super().__init__()
        self.loaders = loaders
        self.parent = parent
        self.base_dir = base_dir

        self.state = initial_state
        if self.state is None:
            self.state = LoadDataDialogState()

        self.ui = LoadWindow()
        self.setWindowIcon(parent.windowIcon())
        self.ui.setupUi(self)
        self.setStyleSheet(parent.styleSheet())
        self._setup_ui()
        self._init_position()

    def _init_position(self):
        """Move the window to the center of the parent window."""

        x = self.parent.pos().x() + self.parent.size().width() / 2 - self.size().width() / 2
        y = self.parent.pos().y() + self.parent.size().height() / 2 - self.size().height() / 2
        self.move(x, y)

    def _setup_ui(self):
        self.setWindowTitle("Load Data")
        self.ui.combo_plugin.addItems([loader.name() for loader in self.loaders])

        self.ui.btn_select_data.clicked.connect(lambda: self._handle_file_select("data_file"))
        self.ui.btn_select_video.clicked.connect(lambda: self._handle_file_select("video_file"))
        self.ui.btn_select_annotation.clicked.connect(lambda: self._handle_file_select("annotation_file"))

        self.state.bind_bidirectional(self.ui.qedit_data_path.setText, self.ui.qedit_data_path.textEdited, "data_file")
        self.state.bind_bidirectional(
            self.ui.qedit_video_path.setText, self.ui.qedit_video_path.textEdited, "video_file"
        )
        self.state.bind_bidirectional(
            self.ui.qedit_annotation_path.setText, self.ui.qedit_video_path.textEdited, "annotation_file"
        )

        self.ui.btn_ok.clicked.connect(self._handle_ok_click)
        self.ui.btn_cancel.clicked.connect(self.close)

        light = Config.theme.COLOR_LIGHT

        for label in self.findChildren(QLabel):
            label.setStyleSheet(f"color: rgb({light.red()},{light.green()},{light.blue()});")

        for edit in self.findChildren(QLineEdit):
            edit.setStyleSheet(f"color: rgb({light.red()},{light.green()},{light.blue()});")

        for elem in self.findChildren(QPushButton):
            elem.setStyleSheet(self.parent.ui.btn_add_label.styleSheet())

        style_cb = self.parent.ui.btn_add_label.styleSheet().replace("QPushButton", "QComboBox")
        self.ui.combo_plugin.setStyleSheet(style_cb)
        self.ui.combo_plugin.view().setStyleSheet(style_cb.replace("QComboBox", "QListView"))

    def _handle_file_select(self, property_name):
        loader = self.loaders[self.ui.combo_plugin.currentIndex()]
        try:
            file_type = loader.file_type[property_name]
        except KeyError:
            # apparently the plugin does not restrict files to be displayed in openfiledialog
            file_type = ""
        file_name = ask_for_file_name(self.base_dir, parent=self, file_type=file_type)
        if file_name is not None:
            self.state.set(property_name, file_name)
            self.base_dir = str(Path(file_name).parent)

    def _handle_ok_click(self):
        """Use the selected loader for the selcted data.

        Additionally, this changes to cursor to `busy` for user feedback while loading the data.
        """
        set_cursor(self, QtCore.Qt.BusyCursor)
        final_data, loader = self._process_data()
        set_cursor(self, QtCore.Qt.ArrowCursor)
        if final_data is None or loader is None:
            return
        self.final_data_ = final_data

        self.accept()

    def _process_data(self):
        if not self.state.data_file:
            UserInformation().inform("You need to select a sensor data file!")
            return None, None

        # Get loader from combobox
        self.loader_ = self.loaders[self.ui.combo_plugin.currentIndex()]
        try:
            # TODO: Implement loader config
            user_config = {}
            loader = self.loader_(parent=self, **user_config)
        except Exception as e:  # noqa
            # ignore bare except because anything can go wrong in a user-implemented plugin
            print(e)
            UserInformation().inform(f"Error creating an instance of the plugin {self.loader_.name}:\n\n {e}")
            return None, None

        try:
            data = loader.load_sensor_data(self.state.data_file)
        except Exception as e:  # noqa
            self.setCursor(Qt.ArrowCursor)
            UserInformation.inform(
                f"There was an error loading the data ({str(e)}). Maybe you selected a wrong file or a wrong "
                f"recording system in the dropdown box?"
                f"\n\n"
                f"For the complete error message, see the terminal.\n"
            )
            warnings.warn(traceback.format_exc())
            return None, None

        self.validate_data_format(data)

        if self.state.annotation_file:
            annotations = loader.load_annotations(self.state.annotation_file)
            data = self._incorporate_annotations_to_data(data, annotations)

        return_dict = {
            "plot_data_dicts": data,
            "data_file_name": self.state.data_file,
            "start_time": loader.get_start_time(self.state.data_file),
        }

        if self.state.video_file:
            # Note: this must be done after loading data, since loading video might trigger plotting yellow lines in
            # the plots for video-signal-synchronization
            return_dict = self._handle_video_file(return_dict, loader)

        return return_dict, loader

    def validate_data_format(self, plot_data: Dict):
        if not isinstance(plot_data, dict):
            UserInformation.inform(
                "The loader's `load_sensor_data` method must return a dict, where the keys are the names you would "
                "like to give the plots. "
                "Click `Learn More` for more information.",
                help_link=LINK_IMPLEMENT_IMPORTER,
            )
            raise KeyError(
                f"{self.loader_.name()}'s  `load_sensor_data` method must return a dict. See {LINK_IMPLEMENT_IMPORTER}"
            )

        for plot, data in plot_data.items():
            if "sensor_data" not in data.keys():
                UserInformation.inform(
                    "Missing key `sensor_data` in the nested dict returned by the loader. See console output for "
                    "more information or click the link below to see what the importer should return",
                    help_link=LINK_IMPLEMENT_IMPORTER,
                )
                raise KeyError(
                    f"{self.loader_.name()} returned data to be plotted with the name {plot}. {plot} does not contain a"
                    f" key `sensor_data`, but is expected to. `sensor_data` in turn should keep a pd.DataFrame, "
                    f"where the columns are the channels to plot and each row is one sample to plot. "
                    f"See {LINK_IMPLEMENT_IMPORTER}"
                )

            if "sampling_rate_hz" not in data.keys():
                UserInformation.inform(
                    "Missing key `sampling_rate_hz` in the nested dict returned by the loader. See console output for "
                    "more information or click the link below to see what the importer should return",
                    help_link=LINK_IMPLEMENT_IMPORTER,
                )
                raise KeyError(
                    f"{self.loader_.name()} returned data to be plotted with the name {plot}. {plot} does not contain a"
                    f" key `sampling_rate_hz`, but is expected to. `sampling_rate_hz` in turn should keep a float. "
                    f"See {LINK_IMPLEMENT_IMPORTER}"
                )
            sensor_data = data.get("sensor_data", None)

            if not isinstance(sensor_data, pd.DataFrame):
                UserInformation.inform(
                    text="Data format was not valid. If you are a developer, see the command line for more detailed "
                    "information or click `Learn More` to get to our documentation about importers.",
                    help_link=LINK_IMPLEMENT_IMPORTER,
                )
                raise KeyError(
                    f"You tried to load data named {plot} using {self.loader_.name()}'s `load_sensor_data`. However, "
                    f"the key `sensor_data` keeps data of the type {type(sensor_data)}, although it should be a "
                    f"pandas DataFrame. See {LINK_IMPLEMENT_IMPORTER} for more info."
                )

    @staticmethod
    def _incorporate_annotations_to_data(data: Dict, annotations: Dict) -> Dict:
        for sensor, annotation in annotations.items():
            try:
                data[sensor]["annotations"] = annotation
            except KeyError as k:
                UserInformation.inform(
                    "Loader provided annotations for sensors that have no plot. Click 'Learn More' "
                    "for more information",
                    help_link="https://mad-gui.readthedocs.io/en/latest/"
                    "troubleshooting.html#loader-provided-"
                    "annotations-that-were-not-understood",
                )

                raise ValueError(
                    "The dict keys of the annotations must match the dict keys of the sensor data. See "
                    "https://mad-gui.readthedocs.io/en/latest/troubleshooting.html#id2 for more "
                    "information."
                ) from k
        return data

    def _handle_video_file(self, return_dict, loader):
        return_dict["video_file"] = self.state.video_file
        sync_file = loader.get_sync_file(self.state.video_file)
        if sync_file:
            return_dict["sync_file"] = sync_file
        return return_dict

    def get_data(self) -> Optional[Tuple[Dict[str, Dict[str, Any]], BaseImporter]]:
        """Run this dialog and return the data, that was selected by the user."""
        if self.exec_():
            return self.final_data_, self.loader_
        return None, None



# pylint: skip-file
"""
MaD GUI
"""

from mad_gui.config import BaseSettings, BaseTheme
from mad_gui.plugins import BaseImporter, BaseExporter, BaseAlgorithm
from mad_gui.start_gui import start_gui

__all__ = ["BaseSettings", "BaseTheme", "start_gui", "BaseImporter", "BaseExporter", "BaseAlgorithm"]



"""A window to display a message for the user and potentially getting a yes/no answer."""
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QDialog, QMessageBox

from typing import Optional


class UserInformation(QDialog):
    """A dialog to transport information to/from user.

    Methods
    -------
    inform
        Send a message to the use to be accepted with the `OK` button.
    confirm
        Pose a yes/no question and obtain the answer.
    """

    @classmethod
    def _create_message(
        cls, text: str, buttons: [QMessageBox.StandardButton], help_link: Optional[str] = None
    ) -> QMessageBox:
        """Used in different contexts to create a message

        Parameters
        ----------
        text
            The text you want to show the user
        help_link
            A URL-link to supplementary information
        buttons
            The buttons are the options the user has to answer your text
        """
        msg = QMessageBox()
        for i_button in buttons:
            msg.addButton(i_button)
        if help_link:
            text = text + (
                ' <html><head/><body><p><span style=" color:#063d69;"></span><a href="'
                + help_link
                + '"><span style=" text-decoration: underline; color:#063d69;">Learn More'
            )
        msg.setText(text)
        msg.setAutoFillBackground(True)
        msg.setWindowTitle("Information")
        msg.setWindowFlag(Qt.FramelessWindowHint, True)
        return msg

    @classmethod
    def inform(cls, text: str, help_link: Optional[str] = None) -> bool:
        """Send a message to the user in a pop-up window with an OK-Button.

        For example this might be useful if something crashed during loading data or executing an algoritihm and you
        would like to provide more context to the user.

        Parameters
        ----------
        text
            The information you want to give to the user.

        help_link
            If you pass this, it will hide behind the `Learn More` button in the lower part of the message.



        Examples
        --------
        Note: If you want to execute the following example without having a MaD GUI instance running (i.e. you did
        not use :meth:`mad_gui.start_gui`, you need to:

        >>> from PySide2.QtWidgets import QApplication
        >>> app = QApplication()

        You do not need the above code snippet, if you want to implement the lines below into your plugin.

        >>> from mad_gui.components.dialogs import UserInformation
        >>> UserInformation.inform("Please make sure to X")
        """
        if help_link:
            msg = cls._create_message(text, [QMessageBox.StandardButton.Ok], help_link)
        else:
            msg = cls._create_message(text, [QMessageBox.StandardButton.Ok])
        msg.exec_()

    @classmethod
    def confirm(cls, text: str, help_link: Optional[str] = None) -> bool:
        """Ask the user something and obtain yes or no as answer.

        Parameters
        ----------
        text
            The question you want to pose to the user.

        help_link
            If you pass this, it will hide behind the `Learn More` button in the lower part of the message.

        Returns
        -------
        answer
            Either `Yes` or `No` of the class :py:class:`PySide2.QtWidgets.QMessageBox`.
            :py:

        Examples
        --------
        Note: If you want to execute the following example without having a MaD GUI instance running (i.e. you did
        not use :meth:`mad_gui.start_gui`, you need to:

        >>> from PySide2.QtWidgets import QApplication
        >>> app = QApplication()

        You do not need the above code snippet, if you want to implement the lines below into your plugin.

        >>> from PySide2.QtWidgets import QMessageBox
        >>> from mad_gui.components.dialogs import UserInformation
        >>> answer = UserInformation.confirm("Do you want to perform X?")
        >>> if answer == QMessageBox.Yes:
        ...    print("Yes!")
        ...
        Yes!
        """
        msg = cls()._create_message(text, [QMessageBox.Yes, QMessageBox.No], help_link)
        return msg.exec_()



from PySide2.QtCore import QObject, Qt, Slot
from PySide2.QtGui import QKeyEvent

from mad_gui.models.ui_state import PlotState
from typing import Optional, Tuple


def move_range(old_range: Tuple[float, float], distance: float, max_range: Tuple[float, float]) -> Tuple[float, float]:
    # TODO: Move to global helper
    if old_range[0] + distance < max_range[0]:
        return max_range[0], old_range[1] + (max_range[0] - old_range[0])
    if old_range[1] + distance > max_range[1]:
        return old_range[0] + (max_range[1] - old_range[1]), max_range[1]
    return old_range[0] + distance, old_range[1] + distance


class KeyEventHandler(QObject):
    """Used as a singleton to propagate key press events from children to :class:`mad_gui.windows.MainWindow`."""

    STATE_CHANGE = {
        Qt.Key_A: "add",
        Qt.Key_E: "edit",
        Qt.Key_R: "remove",
        Qt.Key_S: "sync",
        Qt.Key_I: "investigate",
        Qt.Key_Escape: "investigate",
    }

    def __init__(self, plot_state: PlotState, parent: Optional[QObject] = None):
        super().__init__(parent)
        self.plot_state = plot_state

    @Slot(QKeyEvent)
    def key_pressed(self, event: QKeyEvent):
        self._global_mode_change_events(event)
        self._global_plot_move(event)

    def _global_mode_change_events(self, ev):
        new_mode = self.STATE_CHANGE.get(ev.key(), None)
        if new_mode:
            ev.accept()
            if new_mode == self.plot_state.mode:
                self.plot_state.mode = "investigate"
                return
            self.plot_state.mode = new_mode

    def _global_plot_move(self, ev):
        if ev.key() == Qt.Key_Q:
            old_range = self.plot_state.x_range
            jump_range = (old_range[1] - old_range[0]) * 0.8
            if ev.modifiers() == Qt.ShiftModifier:
                # move backwards
                self.plot_state.x_range = move_range(old_range, -jump_range, self.plot_state.x_range_max)
            else:
                # move forward
                self.plot_state.x_range = move_range(old_range, jump_range, self.plot_state.x_range_max)
            ev.accept()



from PySide2 import QtCore
from PySide2.QtUiTools import loadUiType
from PySide2.QtWidgets import QDialog

from mad_gui.components.dialogs.user_information import UserInformation
from mad_gui.components.helper import set_cursor
from mad_gui.models.global_data import GlobalData
from mad_gui.models.local import PlotData
from mad_gui.plugins.base import BasePlugin
from mad_gui.qt_designer import UI_PATH
from mad_gui.utils.helper import resource_path
from typing import List, Type

ui_path = resource_path(str(UI_PATH / "plugin_selection.ui"))
if ".ui" in ui_path:
    try:
        UiForm, _ = loadUiType(ui_path)
    except TypeError:
        try:
            import sys
            import os
            from pathlib import Path

            uic_path = Path(os.sep.join(sys.executable.split(os.sep)[:-1])) / "Scripts"
            sys.path.append(str(uic_path))
            UiForm, _ = loadUiType(ui_path)
        except TypeError as e:
            raise FileNotFoundError(
                "Probably python did not find `pyside2-uic`. See "
                '"https://mad-gui.readthedocs.io/en/latest/troubleshooting.html#pyside2-uic-not-found" for more '
                "information"
            ) from e

elif ".py" in ui_path:
    from mad_gui.qt_designer.build.plugin_selection import Ui_Form as UiForm  # noqa


class PluginSelectionDialog(QDialog):
    """A dialog to select the plugin to use.

    See Also
    --------
    :class:`~mad_gui.windows.BasePluginSelector`

    """

    _data: PlotData

    def __init__(self, plugins: List[Type[BasePlugin]], parent=None):
        super().__init__()
        self.plugins = plugins
        self.parent = parent
        self.ui = UiForm()
        self.setWindowIcon(parent.windowIcon())
        self.ui.setupUi(self)
        self.setStyleSheet(parent.styleSheet())
        self._setup_ui()
        self.executed_plugin = None

    def _setup_ui(self):
        self.setWindowTitle("Select Plugin")
        self.ui.combo_plugin.addItems([plugin.name() for plugin in self.plugins])
        self.ui.btn_ok.clicked.connect(self._start_processing)
        self.ui.btn_cancel.clicked.connect(self.close)

        for btn in [
            self.ui.btn_ok,
            self.ui.btn_cancel,
        ]:
            btn.setStyleSheet(self.parent.ui.btn_add_label.styleSheet())

        style_cb = self.parent.ui.btn_add_label.styleSheet().replace("QPushButton", "QComboBox")
        self.ui.combo_plugin.setStyleSheet(style_cb)
        self.ui.combo_plugin.view().setStyleSheet(style_cb.replace("QComboBox", "QListView"))

    def _start_processing(self):
        set_cursor(self, QtCore.Qt.BusyCursor)
        out = self._process_data()
        set_cursor(self, QtCore.Qt.ArrowCursor)
        if not out:
            return
        self.accept()

    def _process_data(self):
        """Use the selected plugin to create a dataframe to be exported into an excel file.

        The emitted signal will be fetched by :class:`mad_gui.windows.MainWindow`, which will in turn trigger the
        loading by the regarding importer. Also, the mode of the GUI will be set to `investigate`.
        """
        try:
            plugin_class = self.plugins[self.ui.combo_plugin.currentIndex()]
        except IndexError:
            UserInformation.inform("No methods for exporting are implemented.")
            return False
        try:
            # TODO: Implement loader config
            user_config = {}
            plugin = plugin_class(parent=self, **user_config)
        except Exception as error:  # pylint: disable=broad-except
            # broad exception on purpose because we do not know which exceptions might be thrown by an plugin
            # created by someone else
            UserInformation().inform(f"Error loading Plugin {plugin_class.name()} \n Error:\n {str(error)}")
            return False

        try:
            plugin.process_data(self._data)
            self.executed_plugin = plugin_class
        except Exception as error:
            UserInformation().inform(
                f"An error occured inside your plugin {plugin_class.name()}: {str(error)}\n"
                f"Try to debug by setting a breakpoint in your plugin {plugin_class.name()} or see the command line "
                f"for more information."
            )
            raise error
        return True

    def process_data(self, data: GlobalData):
        self._data = data
        self.exec_()



from mad_gui.components.dialogs.user_information import UserInformation
from mad_gui.components.dialogs.label_annotation_dialog import NestedLabelSelectDialog

__all__ = ["UserInformation", "NestedLabelSelectDialog"]



"""This builds a GUI which can
a) load and show IMU data
b) apply an algorithm for stride segmentation and event detection
c) be used to manually add/delete/adapt labels for strides and/or activites.

isort:skip_file (Required import order: PySide2, pyqtgraph, mad_gui.*)
"""
import os
import sys
import warnings
from pathlib import Path
import platform
import ctypes
import pickle
from typing import Dict, Tuple, List

import pandas as pd
import pyqtgraph as pg
from PySide2.QtCore import Qt
from PySide2.QtUiTools import loadUiType
from PySide2.QtWidgets import (
    QFileDialog,
    QMessageBox,
    QVBoxLayout,
    QMainWindow,
    QApplication,
)
from PySide2.QtGui import QPalette

from mad_gui.components.dialogs.data_selector import DataSelector
from mad_gui.components.dialogs.plugin_selection.load_data_dialog import LoadDataDialog
from mad_gui.components.dialogs.plugin_selection.plugin_selection_dialog import PluginSelectionDialog
from mad_gui.components.dialogs.user_information import UserInformation
from mad_gui.components.helper import set_cursor
from mad_gui.components.key_event_handler import KeyEventHandler
from mad_gui.components.sidebar import Sidebar
from mad_gui.config import Config, BaseSettings, BaseTheme
from mad_gui.models.global_data import GlobalData
from mad_gui.models.local import PlotData
from mad_gui.models.ui_state import UiState, PlotState, MODES
from mad_gui.plot_tools.plots import SensorPlot, VideoPlot
from mad_gui.plot_tools.labels import BaseRegionLabel, BaseEventLabel
from mad_gui.plugins.base import BaseExporter, BaseImporter, BaseAlgorithm
from mad_gui.plugins.helper import filter_plugins
from mad_gui.state_keeper import StateKeeper
from mad_gui.utils.helper import resource_path
from mad_gui.windows import VideoWindow
from mad_gui.qt_designer import UI_PATH

try:
    import pyi_splash  # noqa

    pyi_splash.close()
except ModuleNotFoundError:
    # we only need to import this when we are in a .exe, see pyinstaller docs
    pass

# helps to make plot zooming smooth even when line width >1
pg.setConfigOption("useOpenGL", True)

# CI can't handle openGL
if os.environ.get("GITHUB_CI"):
    pg.setConfigOption("useOpenGL", False)

# Make sure that graphs are properly scaled when having multiple screens
QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
if platform.system() == "Windows" and int(float(platform.release())) >= 8:
    # we need int(float(...)) because if issues with windows 8
    ctypes.windll.shcore.SetProcessDpiAwareness(True)

if platform.system() == "Darwin":
    os.environ["QT_MAC_WANTS_LAYER"] = "1"

ui_path = resource_path(str(UI_PATH / "main.ui"))
if ".ui" in ui_path:
    try:
        Window, _ = loadUiType(ui_path)
    except TypeError:
        try:
            uic_path = Path(os.sep.join(sys.executable.split(os.sep)[:-1])) / "Scripts"
            sys.path.append(str(uic_path))
            Window, _ = loadUiType(ui_path)
        except TypeError as error:
            raise FileNotFoundError(
                "Probably python did not find `pyside2-uic`. See "
                '"https://mad-gui.readthedocs.io/en/latest/troubleshooting.html#pyside2-uic-not-found" for more '
                "information"
            ) from error
elif ".py" in ui_path:
    from mad_gui.qt_designer.build.main import Ui_MainWindow as Window  # noqa


class MainWindow(QMainWindow):
    """This class implements the functionalities of the buttons in the GUI.
    Furthermore, it serves as an interface to Input-Output files, which are different for each data source,
    see our `General Information` part of the docs, section `Adding support for other systems`."""

    def __init__(
        self,
        parent=None,
        data_dir=None,
        settings=BaseSettings,
        theme=BaseTheme,
        plugins=None,
        labels=None,
        events=None,
    ):

        super().__init__()

        self.check_arguments(plugins, labels, events)

        Config.set_theme(theme)
        Config.set_settings(settings)
        self.global_data = GlobalData(parent=self)
        self.ui_state = UiState(parent=self)
        self.plot_state = PlotState(parent=self)
        self.global_data.labels = labels or []
        self.global_data.events = events or []

        self.parent = parent

        # Setting up the UI
        self.ui = Window()
        self.ui.setupUi(self)
        c = theme.COLOR_DARK

        self.setStyleSheet(f"background-color: rgb({c.red()}, {c.green()}, {c.blue()});")
        self.setStyleSheet(f"background-color: rgb({c.red()}, {c.green()}, {c.blue()});")
        self.palette().setColor(QPalette.Active, QPalette.Window, theme.COLOR_LIGHT)
        self._set_window_properties()

        # Register sidebar component logic
        self.menu = Sidebar(self.ui, parent=parent)
        self.ui_state.bind_bidirectional(self.menu.set_collapsed, self.menu.collapsed_changed, "menu_collapsed")
        self.menu.set_collapsed(False)

        # Setting up additional windows
        self.VideoWindow = VideoWindow(parent=self)
        self.data_selector = None

        # can only be done after adding the windows above
        self.label_buttons = {
            "add": self.ui.btn_add_label,
            "edit": self.ui.btn_edit_label,
            "remove": self.ui.btn_remove_label,
            "sync": self.ui.btn_sync_data,
        }

        self.menu_buttons = {
            "load": self.ui.btn_load_data,
            "algorithm": self.ui.btn_use_algorithm,
            "export": self.ui.btn_export,
            "save": self.ui.btn_save_data_gui_format,
        }

        # Setting up the plots
        self.sensor_plots = {}
        self.video_plot = None

        self._configure_buttons()

        # Setting up some attributes
        self.data_types = None  # data that the user wants to load/save (sensor, activities, and/or strides)
        self._user_is_informed = False

        self.closeEvent = self._close_event  # doing this to have consistent method naming

        self.global_data.bind(self.ui.label_displayed_data.setText, "data_file")
        self.global_data.bind(self._plot_data, "plot_data", initial_set=False)
        self.global_data.bind(self._set_sync, "sync_file", initial_set=False)
        self.plot_state.bind(self._update_button_state, "mode", initial_set=True)

        # Setup Key Event handler
        self.key_event_handler = KeyEventHandler(plot_state=self.plot_state, parent=self)
        self.keyPressEvent = self.key_event_handler.key_pressed

        StateKeeper.setParent(self)
        StateKeeper.announce_data_types.connect(self._set_data_types)
        StateKeeper.save_sync.connect(self._save_sync)

        # Note: Need to make all connections and ui setup before updating the value
        self.global_data.base_dir = data_dir
        self.global_data.plugins = list(plugins)

    def check_arguments(self, plugins, labels, events):
        for plugin in plugins:
            self._check_argument(plugin, (BaseImporter, BaseAlgorithm, BaseExporter))

        for label in labels:
            self._check_argument(label, (BaseRegionLabel,))
            if label.min_height > label.max_height:
                raise ValueError(
                    f"For the class {label.__name__}, min_height is higher than max_height, please fix that."
                )
            if label.max_height > 1:
                raise ValueError(
                    f"For the class {label.__name__}, max_height is > 1, although it should be between 0 "
                    f"and 1. Please fix that."
                )

        for event in events:
            self._check_argument(event, (BaseEventLabel,))

    @staticmethod
    def _get_element_base(plugin):
        if issubclass(plugin, BaseRegionLabel):
            return "labels"
        if issubclass(plugin, BaseEventLabel):
            return "events"
        if issubclass(plugin, (BaseImporter, BaseAlgorithm, BaseExporter)):
            return "plugin"
        return "unknown"

    def _check_argument(self, element, base_classes: Tuple):
        if not issubclass(element, base_classes):
            base = self._get_element_base(element)
            if base == "unknown":
                raise ValueError(
                    f"{element.__name__} must inherit from one of BaseImporter, BaseAlgorithm, "
                    f"BaseExporter, BaseRegionLabel, or BaseEventLabel but it does not."
                )
            raise ValueError(
                f"You passed {element} with the keyword 'plugin' to the GUI. However, "
                f"your plugin does not inherit from BaseImporter, BaseAlgorithm, or BaseExporter.\n"
                f"You should have passed it with: start_gui({base}=[{element.__name__}])"
            )

    def _enable_buttons(self, enable: bool):
        """In the beginning we want the user to load data, so we just show the two buttons."""
        for button in [
            self.ui.btn_add_label,
            self.ui.btn_edit_label,
            self.ui.btn_remove_label,
            self.ui.btn_sync_data,
            self.ui.btn_export,
            self.ui.btn_save_data_gui_format,
            self.ui.btn_use_algorithm,
        ]:
            if len(self.sensor_plots) == 1 and button == self.ui.btn_sync_data:
                continue
            button.setDisabled(not enable)

    def is_data_plotted(self):
        return bool(self.sensor_plots)

    def _configure_buttons(self):
        # buttons menu
        self.ui.btn_use_algorithm.clicked.connect(self.use_algorithm)
        self.ui.btn_load_data.clicked.connect(self.import_data)
        self.ui.btn_save_data_gui_format.clicked.connect(self.save_data_gui_format)
        self.ui.btn_export.clicked.connect(self.export)
        self.ui.btn_load_data_gui_format.clicked.connect(self._handle_load_data_gui_format)
        # buttons manual annotation
        light = Config.theme.COLOR_LIGHT
        dark = Config.theme.COLOR_DARK
        light_hsl = light.toHsl()
        even_lighter = light_hsl.lighter(150).toRgb()
        for k, b in self.label_buttons.items():
            b.setObjectName(k)
            b.toggled.connect(self.on_main_buttons_clicked)
            qt_light = f"rgb({light.red()}, {light.green()}, {light.blue()})"
            qt_lighter = f"rgb({even_lighter.red()}, {even_lighter.green()}, {even_lighter.blue()})"
            qt_dark = f"rgb({dark.red()}, {dark.green()}, {dark.blue()})"
            b.setStyleSheet(
                f"QPushButton"
                f"{{\nborder:2px solid;\nborder-color: {qt_light};\npadding: 3px;\nbackground-color:{qt_light};"
                f"\ntext-align: left;\ncolor:{qt_dark};\n"
                f"border-radius: 5px;}}\n\n"
                f"QPushButton:hover{{\nborder:2px solid;\nborder-color: {qt_lighter};\n}}"
                f"QPushButton:disabled{{\n	background-color: rgb(160,160,160);\n"
                f"color: rgb(120,120,120)}}"
                f"QPushButton:checked{{background-color: {qt_lighter};\n}}"
                f"QPushButton:pressed{{\n	background-color: {qt_lighter};\n}}"
            )
            b.setFlat(False)
        for k, b in self.menu_buttons.items():
            b.setObjectName(k)
            b.toggled.connect(self.on_main_buttons_clicked)
            b.setStyleSheet(
                f"QPushButton"
                f"{{\nborder:none;\npadding: 3px;\nbackground-color:rgb({light.red()},{light.green()},{light.blue()});"
                f"\ntext-align: left;\ncolor:rgb({dark.red()},{dark.green()},{dark.blue()});\n"
                f"border-radius: 0px;}}\n\n"  # to remove shadow around button
                f"QPushButton:hover{{\n	background-color: rgb("
                f"{even_lighter.red()},{even_lighter.green()},{even_lighter.blue()});\n}}"
                f"QPushButton:disabled{{\n"
                f"color: rgb(120,120,120)}}"
                f"QPushButton:checked{{\n	background-color: rgb("
                f"{even_lighter.red()},{even_lighter.green()},{even_lighter.blue()});\n}}"
            )
        self._enable_buttons(False)

    def on_main_buttons_clicked(self, new_state):
        button = self.sender()
        if new_state is False:
            self.plot_state.mode = "investigate"
            return
        self.plot_state.mode = button.objectName()

    def _update_button_state(self, new_mode: MODES):
        for k, b in self.label_buttons.items():
            if not b.isEnabled():
                return
            old_state = b.blockSignals(True)
            b.setChecked(k == new_mode)
            b.blockSignals(old_state)

    def _unlink_plots(self):
        for plot in [*self.sensor_plots, self.ui.video_plot]:
            plot.set_coupled_plot(None)

    def _save_sync(self):
        sync = pd.DataFrame()
        main_plot = self._get_main_plot()
        if getattr(Config.settings, "SENSORS_SYNCHRONIZED", False):
            for plot in self.sensor_plots.values():
                if plot == main_plot:
                    continue
                plot.sync_info = main_plot.sync_info
        self._link_plots()

        for plot_name, plot in self.sensor_plots.items():
            sync = pd.concat([sync, pd.DataFrame(data=plot.sync_info, columns=[plot_name + "_sample"])], axis=1)
        sync = pd.concat([sync, pd.DataFrame(data=self.video_plot.sync_info, columns=["video_ms"])], axis=1)
        self.VideoWindow.set_sync(self.video_plot.sync_info["start"], self.video_plot.sync_info["end"])
        file_name, _ = QFileDialog.getSaveFileName(self, "Save Synchronization File", filter="*.xlsx")
        if file_name is None:
            return
        sync.to_excel(file_name)
        for plot in self.sensor_plots.values():
            plot.adapt_to_opening_video_window()

    def _handle_load_data_gui_format(self):
        if self.is_data_plotted():
            answer = UserInformation.confirm("Plotted data might be overwritten. Do you want to continue?")
        else:
            answer = QMessageBox.Yes

        if answer == QMessageBox.Yes:
            file = self._ask_for_file_name(file_type="*.mad_gui")
            self.load_data_from_pickle(file)

    def _ask_for_file_name(self, file_type=None):
        data_file = QFileDialog().getOpenFileName(
            parent=None, caption="Select file to open", dir=str(self.global_data.base_dir), filter=file_type
        )[0]
        if data_file == "":
            # User clicked cancel
            return None
        self.global_data.data_file = data_file
        return data_file

    def _set_data_types(self, types: Dict):
        """Saves which of [sensor data, activity labels, stride labels] the user wants to load/save."""
        self.data_types = types

    def _parse_labels_to_load(self, plot_data: Dict):
        label_classes_to_load = []
        for sensor_item in plot_data.values():
            if "annotations" in sensor_item.keys():
                label_classes_to_load.extend(sensor_item["annotations"].keys())

        loadable_labels = []
        unknown_labels = []
        labels_and_events = (*self.global_data.labels, *self.global_data.events)
        known_label_types = {label.name: label for label in labels_and_events}

        for label_class in label_classes_to_load:
            if label_class in known_label_types.keys():
                loadable_labels.append(label_class)
            else:
                unknown_labels.append(label_class)

        if len(unknown_labels) > 0:
            UserInformation.inform(
                f"The saved data has labels which this GUI does not know.\n\n"
                f"Unknown label class: {unknown_labels}\n"
            )
            warnings.warn("Implement link to help.")
        return loadable_labels

    def load_data_from_pickle(self, file: str):
        """Load data from a .mad_gui file.

        Parameters
        ----------
        file
            Full path to to a file that ends with `.mad_gui`. This file was previously created using
            :func:`~mad_gui.MainWindow.save_data_gui_format` and contains sensor data, activity labels and stride
            labels.
            However, the user might previously have selected that not all of those should be loaded.
            Which parts should be loaded is stored in `self.data_types`.
        """
        if file is None:
            # user clicked abort
            return None, None
        if file.split(".")[-1] != "mad_gui":
            UserInformation.inform("Can only load files that end with '.mad_gui'.")
            return None, None
        self.setCursor(Qt.BusyCursor)
        loaded_data = pd.read_pickle(file)

        loadable_labels = self._parse_labels_to_load(loaded_data)

        # Doing it in two lines, and exposing via self to enable testing this whole method
        self.data_selector = DataSelector(parent=self, labels=set(loadable_labels))
        self.data_selector.ask_user()

        if not self.data_types["sensor_data"] and not self.global_data.plot_data:
            UserInformation.inform(
                "Can only plot labels if data is plotted. Please also tick 'Sensor data' or "
                "load sensor data using the 'Load Data' button on the upper left."
            )
            self.setCursor(Qt.ArrowCursor)
            return None, None

        selected_data = [data_type for data_type, use in self.data_types.items() if use]

        plot_data = {}
        for plot_name, data in loaded_data.items():
            plot_data[plot_name] = PlotData.from_dict(data, selections=selected_data)

        self.global_data.plot_data = plot_data
        self.global_data.base_dir = Path(file).parent
        #        self._plot_data()

        self.setCursor(Qt.ArrowCursor)
        self._enable_buttons(True)
        # self.menu.set_collapsed(True)

        return loaded_data, loadable_labels

    def _label_classes_backwards_compatibility(self, labels, known_label_types: Dict):
        class StrideLabel(BaseRegionLabel):
            name = "Stride Label"
            min_height = 0.25
            max_height = 0.75

        class ActivityLabel(BaseRegionLabel):
            name = "Activity Label"
            min_height = 0.75
            max_height = 1

        for label_class in [StrideLabel, ActivityLabel]:
            if label_class.__name__ in labels:
                known_label_types[label_class.__name__] = label_class
                if label_class.__name__ not in [label.__name__ for label in self.global_data.labels]:
                    self.global_data.labels = (*self.global_data.labels, label_class)
        return known_label_types

    def import_data(self):
        """Start dialog to import data.

        This will open a :class:`mad_gui.LoadDataWindow`. In there, the user can select data to be loaded:

            - sensor data
            - annotations
            - video

        Additionally, the user can select the system that was used to record the sensor data and annotations.
        Depending on this selection, the path of the data to be loaded will be passed to the regarding importer in
        :mod:`mad_gui.plugins`.

        """
        loaders = filter_plugins(self.global_data.plugins, BaseImporter)
        if len(loaders) == 0:
            UserInformation.inform(
                "There were no loaders passed to the GUI. Read more about the fact why the plugin you created does "
                "not show up in the GUI by clicking the link below.",
                help_link="https://mad-gui.readthedocs.io/en/latest/troubleshooting.html#the-plugin-i-created-does-not-"
                "show-up-in-the-GUI",
            )
            return

        view = LoadDataDialog(self.global_data.base_dir, loaders=loaders, parent=self)

        data, loader = view.get_data()
        self.global_data.start_time = data["start_time"]

        if data is None:
            return

        self.global_data.active_loader = loader
        self.global_data.data_file = data.get("data_file_name", "")
        self.global_data.sync_file = data.get("sync_file", "")
        self.global_data.video_file = data.get("video_file", "")

        self._parse_labels_to_load(data["plot_data_dicts"])

        try:
            plot_data = {k: PlotData.from_dict(v) for k, v in data["plot_data_dicts"].items()}
        except Exception as e:  # noqa
            raise NotImplementedError(
                f"Possibly there is an error in the data that was loaded using {loader}. Please "
                f"see our guide in implementing an importer: https://mad-gui.readthedocs.io/en/la"
                f"test/customization.html#implement-an-importer"
            ) from e
        self.global_data.plot_data = plot_data
        self.load_video(data.get("video_file", None))
        self._set_sync(data.get("sync_file", None))
        self._enable_buttons(True)
        # self.menu.set_collapsed(True)

    def _set_sync(self, sync_file: str):
        """Set the synchronization for each plot"""
        if not sync_file:
            return
        sync = self.global_data.active_loader.get_video_signal_synchronization(sync_file)

        try:
            plots = [*self.sensor_plots.items(), ("video", self.video_plot)]
        except AttributeError as e:
            print(e)
            plots = self.sensor_plots.items()

        suffixes = {"video": "_ms"}
        unsynced_sensors = []
        for plot_name, plot in plots:
            suffix = suffixes.get(plot_name, "_sample")
            self._set_plot_sync(plot_name, plot, sync[plot_name + suffix], unsynced_sensors)

        if len(unsynced_sensors) > 1:
            UserInformation.inform(
                f"Found a synchronization file ({sync_file}), however there is no info about "
                f"the sensor(s) {unsynced_sensors}, therefore it is not synchronized."
            )

        if self.VideoWindow:
            self.VideoWindow.set_sync(start_frame=sync["video_ms"]["start"], end_frame=sync["video_ms"]["end"])

    @staticmethod
    def _set_plot_sync(plot_name: str, plot, sync: pd.Series, unsynced_sensors: List):
        if plot is None:
            return
        try:
            plot.sync_info = sync
            plot.adapt_to_opening_video_window()
        except KeyError:
            unsynced_sensors.append(plot_name)

    def _plot_data(self, data_dict: Dict[str, PlotData]):

        start_time = getattr(self.global_data, "start_time", 0)
        set_cursor(self, Qt.BusyCursor)

        # Delete all existing plots
        plot_wrapper: QVBoxLayout = self.ui.plotwidget
        for i_plot in list(self.sensor_plots.values()):
            plot_wrapper.removeWidget(i_plot)
            i_plot.deleteLater()
            del i_plot
        self.sensor_plots = {}

        # Create new plots
        for sensor_name, data in data_dict.items():
            plot = SensorPlot(
                plot_data=data,
                initial_plot_channels=getattr(Config.settings, "CHANNELS_TO_PLOT", None),
                start_time=start_time,
                label_classes=self.global_data.labels,
                event_classes=self.global_data.events,
                parent=self,
            )
            plot_wrapper.addWidget(plot)
            self.sensor_plots[sensor_name] = plot
            plot.set_title(sensor_name)
            # Bind global mode change
            self.plot_state.bind_property_bidirectional(plot.state, "mode", "mode", initial="set")

        plots = list(self.sensor_plots.values())
        plots[0].is_main_plot = True

        if getattr(Config.settings, "SENSORS_SYNCHRONIZED", False):
            # Bind the ranges of all plots together
            self._link_plots()
        # TODO: if self.plot_state.mode changes from sync to something else, we want to bind the plots again
        set_cursor(self, Qt.ArrowCursor)

    def _link_plots(self):
        for p in self.sensor_plots.values():
            if p.is_main_plot:
                continue
            self._get_main_plot().set_coupled_plot(p)

    def load_video(self, video_path):
        if not video_path:
            return
        if not os.path.exists(video_path):
            UserInformation.inform(f"The selected file could not be found: {video_path}")
            return
        self.VideoWindow.start_video(video_path)
        self.VideoWindow.show()
        StateKeeper.video_duration_available.connect(self._initialize_video_plot)
        self.ui.btn_sync_data.setDisabled(False)

    def _initialize_video_plot(self):
        StateKeeper.video_duration_available.disconnect(self._initialize_video_plot)
        if self.video_plot:
            self.ui.plotwidget.removeWidget(self.video_plot)
        self.video_plot = VideoPlot(parent=self, video_window=self.VideoWindow)
        self.video_plot.set_title("Video Plot")
        self.video_plot.hide()
        self.ui.plotwidget.addWidget(self.video_plot)
        self.plot_state.bind_property_bidirectional(self.video_plot.state, "mode", "mode", initial="set")
        self.video_plot.sync_info = self.VideoWindow.sync_info

    def _get_main_plot(self):
        return [plot for plot in self.sensor_plots.values() if plot.is_main_plot][0]

    def use_algorithm(self):
        """Applies an algorithm to the plotted IMU data.

        This will basically call :func:`mad_gui.plugins.BaseImporter.annotation_from_data`. Instead of the
        `BaseImporter` a different importer specified in the :class:`mad_gui.LoadDataWindow` (dropdown menu)
        might be used.

        The activity and/or stride labels that will be generated by that method will then be passed to the plots,
        which will subsequently plot the labels, see :func:`mad_gui.plot_tools.SensorPlot.set_activity_labels` and
        :func:`mad_gui.plot_tools.SensorPlot._set_stride_labels`.

        """
        if not self.is_data_plotted():
            UserInformation(parent=self).inform("Please load sensor data before continuing.")
            return

        # if not UserInformation(parent=self).confirm(
        #    "Warning: Calculating new annotations might delete all currently "
        #    "displayed annotations!\nIt is up to the implemented algorithm, "
        #    "if the new annotations are added or if they replace the currently displayed "
        #    "annotations. Do you want to continue?"
        # ):
        #    return

        # Set state to investigate to force updating global state from plot
        self.plot_state.mode = "investigate"

        algorithms = filter_plugins(self.global_data.plugins, BaseAlgorithm)

        if len(algorithms) == 0:
            UserInformation.inform(
                "There were no algorithms passed to the GUI. Read more about the fact why the plugin you created does "
                "not show up in the GUI by clicking the link below.",
                help_link="https://mad-gui.readthedocs.io/en/latest/troubleshooting.html#the-plugin-i-created-does-not-"
                "show-up-in-the-gui",
            )
            return

        set_cursor(self, Qt.BusyCursor)
        dialog = PluginSelectionDialog(plugins=algorithms, parent=self)
        try:
            dialog.process_data(self.global_data.plot_data)
            StateKeeper.executed_algorithms.append(dialog.executed_plugin)
        except Exception as error:  # noqa
            print(sys.exc_info()[0])
            raise NotImplementedError(
                "Possibly there is an error in the implementation of the algorithm. Please "
                "see our guide in implementing an algorithm: https://mad-gui.readthedocs.io/en/latest/customization.ht"
                "ml#implement-an-algorithm"
            ) from error
        self.global_data.plot_data = dialog._data
        set_cursor(self, Qt.ArrowCursor)
        # actually this should be called automatically due to global_data.bind(_plot_data, "plot_data") but that does
        # not work currently
        # we could probably resolve it in mad_gui.plot_tools.plots.SensorPlot in the __init__ but it does not work yet
        self._plot_data(self.global_data.plot_data)

    def _save_data(self, data_to_save: PlotData):
        save_file_name = QFileDialog().getSaveFileName(
            None, "Save GUI data", str(Path(self.global_data.data_file).parent) + "/data.mad_gui", "*.mad_gui"
        )[0]
        if save_file_name != "":
            pickable_data = {k: v.to_dict() for k, v in data_to_save.items()}
            with open(save_file_name, "wb") as file:
                pickle.dump(pickable_data, file, protocol=pickle.HIGHEST_PROTOCOL)

    def save_data_gui_format(self):
        """Saves the displayed sensor data, sampling rate and displayed activity and stride labels into a pickle file.

        The file ending will be `.mad_gui` to make clear it can be loaded again by this GUI. The data can be
        re-loaded into the GUI using the `Load Data GUI format` button.
        If you want to load this data in an other application / script you can so by using :func:`pandas.read_pickle()`.
        """
        if not self.is_data_plotted():
            UserInformation.inform("Please load data before continuing.")
            return

        # Set state to investigate to force updating global state from plot
        self.plot_state.mode = "investigate"

        self._save_data(self.global_data.plot_data)
        StateKeeper.set_has_unsaved_changes(False)

    def export(self):
        """Called when clicking the `Export data` button.

        This button should be used to calculate features from the annotations. For example to calculate a
        mean length of certain activities or strides. To do so, it a :class:`mad_gui.ExportResultsWindow`,
        which basically just is used to select one of the exporters in :mod:`mad_gui.plugins`."""

        if not self.is_data_plotted():
            UserInformation(parent=self).inform("Please to load data before continuing.")
            return

        # Set state to investigate to force updating global state from plot
        self.plot_state.mode = "investigate"

        PluginSelectionDialog(plugins=filter_plugins(self.global_data.plugins, BaseExporter), parent=self).process_data(
            self.global_data
        )

    def _close_event(self, ev):
        if StateKeeper.gui_has_unsaved_changes:
            answer = UserInformation.confirm("Recent changes have not been saved. Are you sure you want to exit?")
            if answer == QMessageBox.No:
                ev.ignore()
                return
        if self.VideoWindow:
            self.VideoWindow.close()
        self.close()

    def _set_window_properties(self):
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setAttribute(Qt.WA_WindowPropagation)
        self.setWindowTitle("MaD GUI")



"""See our Readme -> Adjusting constants for more information"""


class BaseSettings:
    BIND_Y_AXIS = True
    SENSORS_SYNCHRONIZED = True
    AUTO_DOWNSAMPLE = True

    SNAP_RANGE_S = 0.1  # in seconds
    EVENTS = ["Positive peak", "Negative peak"]

    PLOT_WIDTH_PLAYING_VIDEO = 20  # in seconds



import pandas as pd
from PySide2.QtCore import Property as QtProperty
from PySide2.QtCore import QObject, Signal, Slot
from typing_extensions import Literal

from mad_gui.utils import helper
from typing import Any, Callable, Optional, Type, TypeVar


# TODO: Handle nullable values
class PropertyMeta(type(QObject)):
    def __new__(cls, name, bases, attrs):
        for key, attr in list(attrs.items()):
            if not isinstance(attr, Property):
                continue
            initial_value = attr.initial_value
            dtype = attr.dtype
            if not dtype:
                dtype = type(initial_value)
            notifier = Signal(dtype)
            attrs[key] = PropertyImpl(initial_value, name=key, dtype=dtype, notify=notifier)
            attrs[helper.signal_attribute_name(key)] = notifier
        return super().__new__(cls, name, bases, attrs)


T = TypeVar("T")


class Property:
    """Property definition.

    This property will be patched by the PropertyMeta metaclass into a PropertyImpl type.
    """

    def __init__(self, initial_value: Optional[T] = None, dtype: Optional[Type[T]] = None):
        self.initial_value = initial_value
        self.dtype = dtype

    def __get__(self, instance, owner) -> T:
        """Fake getter for typechecker"""
        return self.initial_value

    def __set__(self, instance, value: T):
        """Fake setter for typechecker"""


class PropertyImpl(QtProperty):
    def __init__(self, initial_value, name, dtype=None, notify=None):
        super().__init__(dtype, self._getter, self._setter, notify=notify)
        self.initial_value = initial_value
        self.dtype = dtype
        self.name = name

    def _getter(self, inst):
        return getattr(inst, helper.value_attribute_name(self.name), self.initial_value)

    def _setter(self, inst, value):
        last_value = self._getter(inst)
        if not self._is_equal(last_value, value):
            setattr(inst, helper.value_attribute_name(self.name), value)
            notifier_signal = getattr(inst, helper.signal_attribute_name(self.name))
            notifier_signal.emit(value)

    def _is_equal(self, old, new):
        if self.dtype == pd.DataFrame:
            assert isinstance(old, pd.DataFrame)
            return old.equals(new)
        return old == new


class BaseStateModel(QObject, metaclass=PropertyMeta):
    def bind(self, slot: Slot, property_name: str, initial_set: bool = True):
        helper.set_and_bind_property(slot, self, property_name, initial_set)

    def bind_bidirectional(
        self,
        slot: Slot,
        signal: Signal,
        property_name: str,
        initial_set: bool = True,
        transformer: Optional[Callable] = None,
    ) -> Callable:
        return helper.set_and_bidirectional_bind(slot, signal, self, property_name, initial_set, transformer)

    def bind_property_bidirectional(
        self,
        model: "BaseStateModel",
        target_property_name: str,
        property_name: str,
        initial: Optional[Literal["get", "set"]] = None,
    ):
        if initial == "get":
            target_value = getattr(model, target_property_name)
            setattr(self, property_name, target_value)
        elif initial == "set":
            value = getattr(self, property_name)
            setattr(model, property_name, value)
        return helper.set_and_bidirectional_bind(
            lambda x: setattr(model, target_property_name, x),
            getattr(model, helper.signal_attribute_name(target_property_name)),
            self,
            property_name,
            initial_set=False,
            transformer=None,
        )

    def set(self, property_name: str, value: Any):
        setattr(self, property_name, value)
        return self



from PySide2.QtCore import QObject, QPropertyAnimation, Signal

from mad_gui.config import Config


class Sidebar(QObject):
    """This is not a real sidebar implementation, but just a logic wrapper.

    I would be better to actual use this as widget component, but we would need to change the UI for this.
    """

    def __init__(self, ui, parent=None, initial_collapsed: bool = False):
        super().__init__(parent=parent)
        self.parent = parent
        c = Config.theme.COLOR_LIGHT
        ui.menu_bar.setStyleSheet(f"background-color: rgb({c.red()}, {c.green()}, {c.blue()});")
        self.collapsed = initial_collapsed
        self._animation = None
        self.ui = ui
        self.ui.btn_toggle_menu.clicked.connect(self.toggle)

    def set_collapsed(self, collapsed: bool):
        if collapsed != self.collapsed:
            self.toggle()

    collapsed_changed = Signal(bool)

    def _create_new_animation(self):
        new_animation = QPropertyAnimation(self.ui.menu_middle_frame, b"minimumWidth")
        new_animation.setDuration(150)
        new_animation.setStartValue(self.ui.menu_bar.width())
        new_width = 160 if self.collapsed else 30
        new_animation.setEndValue(new_width)
        return new_animation

    def toggle(self):
        self._animation = self._create_new_animation()
        # self._animation.finished.connect(self._toggle)
        # `finished` would be proper, however, this is never emitted so we use `stateChanged`
        self._animation.stateChanged.connect(self._toggle)
        self._animation.start()

    def _toggle(self):
        if self.collapsed:
            self.ui.btn_load_data.setText("Load data")
            self.ui.btn_export.setText("Export data")
            self.ui.btn_use_algorithm.setText("Use algorithm")
            self.ui.btn_load_data_gui_format.setText("Reload displayed data")
            self.ui.btn_save_data_gui_format.setText("Save displayed data")
            # self.ui.btn_toggle_menu.setText("Collapse menu")
            self.collapsed = False
        else:
            self.ui.btn_load_data_gui_format.setText("")
            self.ui.btn_use_algorithm.setText("")
            self.ui.btn_load_data.setText("")
            self.ui.btn_save_data_gui_format.setText("")
            self.ui.btn_export.setText("")
            # self.ui.btn_toggle_menu.setText("")
            self.collapsed = True
        self.collapsed_changed.emit(self.collapsed)
        self._animation.stateChanged.disconnect(self._toggle)
